#   Copyright (C) 2014 mdm                                     
#   map[dot]plus[dot]plus[dot]help[at]gmail                     
#                                                              
# Licensed to the Apache Software Foundation (ASF) under one   
# or more contributor license agreements.  See the NOTICE file 
# distributed with this work for additional information        
# regarding copyright ownership.  The ASF licenses this file   
# to you under the Apache License, Version 2.0 (the            
# "License"); you may not use this file except in compliance   
# with the License.  You may obtain a copy of the License at   
#                                                              
#   http://www.apache.org/licenses/LICENSE-2.0                 
#                                                              
# Unless required by applicable law or agreed to in writing,   
# software distributed under the License is distributed on an  
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY       
# KIND, either express or implied.  See the License for the    
# specific language governing permissions and limitations            
# under the License.                                             

from __future__ import print_function

import sys
from ctypes import *
import os
import six

from distutils.sysconfig import get_config_var

from sys import platform

libext = get_config_var('EXT_SUFFIX')
if libext is None or libext == '':
    if platform == "linux" or platform == "linux2":
        libext = '.so'
    elif platform == "darwin":
        #libext = '.dyld'
        libext = '.so'
    elif platform == "win32":
        #libext = '.dll'
        libext = '.pyd'
    elif platform == "cygwin":
        libext = '.dll'

maplib = '_libmap' + libext

libpath = os.path.dirname(os.path.abspath(__file__)) + os.path.sep + maplib

"""A collection of methods for the MAP++ program"""
libexec = cdll.LoadLibrary(libpath)

'''
# these are the fortran derived types created by the FAST registry.
f_type_init = None
f_type_initout = None
f_type_d = None
f_type_u = None
f_type_x = None
f_type_y = None
f_type_z = None
f_type_p = None

ierr = c_int(0)
status = create_string_buffer(1024)
summary_file = c_char_p
val = c_double
'''

class ModelData_Type(Structure):
    _fields_ = []


# void * object ;
# double gravity ;
# double seaDensity ;
# double depth ;
# char fileName[255] ;
# char summaryFileName[255] ;
# char libraryInputLine[255] ;
# char nodeInputLine[255] ;
# char elementInputLine[255] ;
# char optionInputLine[255] ;
class InitializationData_Type(Structure):
    _fields_= [("object",c_void_p),
               ("gravity",c_double),
               ("seaDensity",c_double),
               ("depth",c_double),
               ("fileName",c_char*255),
               ("summaryFileName",c_char*255),
               ("libraryInputLine",c_char*255),
               ("nodeInputLine",c_char*255),
               ("elementInputLine",c_char*255),
               ("optionInputLine",c_char*255)]



# void * object ;
# char progName[99] ;
# char version[99] ;
# char compilingData[24] ;
# char * writeOutputHdr ;     int writeOutputHdr_Len ;
# char * writeOutputUnt ;     int writeOutputUnt_Len ;

class InitializationOutputData_Type(Structure):
    _fields_ = [("object",c_void_p),
                ("progName",c_char*99),
                ("version",c_char*99),
                ("CompilingData",c_char*99),
                ("writeOutputHdr",c_char_p),
                ("writeOutputHdr_Len",c_int),
                ("writeOutputUnt",c_char_p),
                ("writeOutputUnt_Len",c_int)]

class InputData_Type(Structure):
    _fields_ = []


class OutputData_Type(Structure):
    _fields_ = []



# void * object ;
# double g ;
# double depth ;
# double rhoSea ;

class ParameterData_Type(Structure):
    _fields_ = [("object",c_void_p),
                ("g",c_double),
                ("depth",c_double), 
                ("rhoSea", c_double)]


class ConstraintData_Type(Structure):
    _fields_ = []


class ContinuousData_Type(Structure):
    _fields_ = []

# fields for the fortran types
# 
# MAP_EXTERNCALL MAP_InitInputType_t* map_create_init_type( char* msg, MAP_ERROR_CODE* status );
# MAP_EXTERNCALL MAP_InitOutputType_t* map_create_initout_type( char* msg, MAP_ERROR_CODE* status );
# MAP_EXTERNCALL MAP_InputType_t* map_create_input_type( char* msg, MAP_ERROR_CODE* status );
# MAP_EXTERNCALL MAP_ParameterType_t* map_create_parameter_type( char* msg, MAP_ERROR_CODE* status );
# MAP_EXTERNCALL MAP_ConstraintStateType_t* map_create_constraint_type( char* msg, MAP_ERROR_CODE* status );
# MAP_EXTERNCALL MAP_OtherStateType_t* map_create_other_type( char* msg, MAP_ERROR_CODE* status );
# MAP_EXTERNCALL MAP_OutputType_t* map_create_output_type( char* msg, MAP_ERROR_CODE* status );
# MAP_EXTERNCALL MAP_ContinuousStateType_t* map_create_continuous_type( char* msg, MAP_ERROR_CODE* status );

MapData_Type       = POINTER(ModelData_Type)
MapInit_Type       = POINTER(InitializationData_Type)
MapInitOut_Type    = POINTER(InitializationOutputData_Type)
MapInput_Type      = POINTER(InputData_Type)
MapOutput_Type     = POINTER(OutputData_Type)
MapParameter_Type  = POINTER(ParameterData_Type)
MapConstraint_Type = POINTER(ConstraintData_Type)
MapContinuous_Type = POINTER(ContinuousData_Type)

# read file stuff
libexec.set_init_to_null.argtype=[MapInit_Type, c_char_p, POINTER(c_int) ]
libexec.map_set_summary_file_name.argtype=[MapInit_Type, c_char_p, POINTER(c_int) ]
libexec.map_add_cable_library_input_text.argtype=[MapInit_Type]
libexec.map_add_node_input_text.argtype=[MapInit_Type]
libexec.map_add_line_input_text.argtype=[MapInit_Type]
libexec.map_add_options_input_text.argtype=[MapInit_Type]

libexec.map_create_init_type.argtype       = [ c_char_p, POINTER(c_int) ]
libexec.map_create_initout_type.argtype    = [ c_char_p, POINTER(c_int) ]
libexec.map_create_input_type.argtype      = [ c_char_p, POINTER(c_int) ]
libexec.map_create_parameter_type.argtype  = [ c_char_p, POINTER(c_int) ]
libexec.map_create_constraint_type.argtype = [ c_char_p, POINTER(c_int) ]
libexec.map_create_other_type.argtype      = [ c_char_p, POINTER(c_int) ]
libexec.map_create_output_type.argtype     = [ c_char_p, POINTER(c_int) ]
libexec.map_create_continuous_type.argtype = [ c_char_p, POINTER(c_int) ]
libexec.map_create_continuous_type.argtype = [ MapData_Type ]

libexec.map_create_init_type.restype       = MapInit_Type
libexec.map_create_initout_type.restype    = MapInitOut_Type
libexec.map_create_input_type.restype      = MapInput_Type
libexec.map_create_parameter_type.restype  = MapParameter_Type
libexec.map_create_constraint_type.restype = MapConstraint_Type
libexec.map_create_other_type.restype      = MapData_Type
libexec.map_create_output_type.restype     = MapOutput_Type
libexec.map_create_continuous_type.restype = MapContinuous_Type


libexec.map_set_sea_depth.argtypes   = [ MapParameter_Type, c_double ]
libexec.map_set_gravity.argtypes     = [ MapParameter_Type, c_double ]
libexec.map_set_sea_density.argtypes = [ MapParameter_Type, c_double ]

libexec.map_size_lines.restype = c_int

# numeric routines
libexec.map_residual_function_length.restype = c_double
libexec.map_residual_function_height.restype = c_double
libexec.map_jacobian_dxdh.restype            = c_double
libexec.map_jacobian_dxdv.restype            = c_double
libexec.map_jacobian_dzdh.restype            = c_double
libexec.map_jacobian_dzdv.restype            = c_double

libexec.map_residual_function_length.argtypes = [ MapData_Type, c_int, c_char_p, POINTER(c_int) ]
libexec.map_residual_function_height.argtypes = [ MapData_Type, c_int, c_char_p, POINTER(c_int) ]
libexec.map_jacobian_dxdh.argtypes            = [ MapData_Type, c_int, c_char_p, POINTER(c_int) ]
libexec.map_jacobian_dxdv.argtypes            = [ MapData_Type, c_int, c_char_p, POINTER(c_int) ]
libexec.map_jacobian_dzdh.argtypes            = [ MapData_Type, c_int, c_char_p, POINTER(c_int) ]
libexec.map_jacobian_dzdv.argtypes            = [ MapData_Type, c_int, c_char_p, POINTER(c_int) ]


libexec.map_get_fairlead_force_2d.argtypes = [POINTER(c_double), POINTER(c_double), MapData_Type, c_int, c_char_p, POINTER(c_int)]
libexec.map_get_anchor_force_2d.argtypes = [POINTER(c_double), POINTER(c_double), MapData_Type, c_int, c_char_p, POINTER(c_int)]

libexec.map_get_fairlead_force_3d.argtypes = [POINTER(c_double), POINTER(c_double), POINTER(c_double), MapData_Type, c_int, c_char_p, POINTER(c_int)]
libexec.map_get_anchor_force_3d.argtypes = [POINTER(c_double), POINTER(c_double), POINTER(c_double), MapData_Type, c_int, c_char_p, POINTER(c_int)]


# plot routines
libexec.map_plot_x_array.argtypes = [ MapData_Type, c_int, c_int, c_char_p, POINTER(c_int) ]
libexec.map_plot_x_array.restype  = POINTER(c_double)
libexec.map_plot_y_array.argtypes = [ MapData_Type, c_int, c_int, c_char_p, POINTER(c_int) ]
libexec.map_plot_y_array.restype  = POINTER(c_double)
libexec.map_plot_z_array.argtypes = [ MapData_Type, c_int, c_int, c_char_p, POINTER(c_int) ]
libexec.map_plot_z_array.restype  = POINTER(c_double)
libexec.map_plot_array_free.argtypes = [ POINTER(c_double) ]



# modifyers
libexec.map_offset_vessel.argtypes = [MapData_Type, MapInput_Type, c_double, c_double, c_double, c_double, c_double, c_double, c_char_p, POINTER(c_int)]        
libexec.map_linearize_matrix.argtypes = [MapInput_Type, MapParameter_Type, MapData_Type, MapOutput_Type, MapConstraint_Type, c_double, POINTER(c_int), c_char_p]        
libexec.map_linearize_matrix.restype  = POINTER(POINTER(c_double))
libexec.map_free_linearize_matrix.argtypes = [POINTER(POINTER(c_double))]

libexec.map_init.argtypes = [ MapInit_Type,
                          MapInput_Type,
                          MapParameter_Type,
                          MapContinuous_Type,
                          MapConstraint_Type,
                          MapData_Type,
                          MapOutput_Type,
                          MapInitOut_Type,
                          POINTER(c_int),
                          c_char_p]


libexec.map_update_states.argtypes = [ c_float,
                                   c_int,
                                   MapInput_Type,
                                   MapParameter_Type,
                                   MapContinuous_Type,
                                   MapConstraint_Type,
                                   MapData_Type,
                                   POINTER(c_int),
                                   c_char_p]

libexec.map_end.argtypes = [ MapInput_Type,
                         MapParameter_Type,
                         MapContinuous_Type,
                         MapConstraint_Type,
                         MapData_Type,
                         MapOutput_Type,
                         POINTER(c_int),
                         c_char_p]


libexec.map_initialize_msqs_base.argtypes = [MapInput_Type,
                                         MapParameter_Type,
                                         MapContinuous_Type,
                                         MapConstraint_Type,
                                         MapData_Type,
                                         MapOutput_Type,
                                         MapInitOut_Type]


libexec.map_size_lines.argtypes = [ MapData_Type,
                                POINTER(c_int),
                                c_char_p]



class pyMAP(object):

    def __init__( self ) :
        self.ierr = c_int(0)
        self.status = create_string_buffer(1024)
        
        self.f_type_d       = self.CreateDataState()
        self.f_type_u       = self.CreateInputState( )
        self.f_type_x       = self.CreateContinuousState( )
        self.f_type_p       = self.CreateParameterState( )
        self.f_type_y       = self.CreateOutputState( )
        self.f_type_z       = self.CreateConstraintState( )
        self.f_type_init    = self.CreateInitState( )
        self.f_type_initout = self.CreateInitoutState( )
        libexec.set_init_to_null(self.f_type_init, self.status, pointer(self.ierr) )
        libexec.map_initialize_msqs_base(self.f_type_u, self.f_type_p, self.f_type_x, self.f_type_z, self.f_type_d, self.f_type_y, self.f_type_initout)
        self.summary_file('outlist.map.sum')


    def init( self ):
        libexec.map_init( self.f_type_init, self.f_type_u, self.f_type_p, self.f_type_x, self.f_type_z, self.f_type_d, self.f_type_y, self.f_type_initout, pointer(self.ierr), self.status )
        if self.ierr.value != 0 : print(self.status.value)


    def size_lines(self):
        size = libexec.map_size_lines(self.f_type_d, pointer(self.ierr), self.status )
        if self.ierr.value != 0 : print(self.status.value)
        return size


    def update_states(self, t, interval):
        libexec.map_update_states(c_float(t), c_int(interval), self.f_type_u, self.f_type_p, self.f_type_x, self.f_type_z, self.f_type_d, pointer(self.ierr), self.status )
        if self.ierr.value != 0 : print(self.status.value)


    # Calls function in main.c and fordatamanager.c to delete insteads of c structs. First, the malloc'ed arrays need to vanish
    # gracefully; we accomplish this by calling MAP_End(...) routine. Then, the structs themself are deleted. Order is important.
    # 
    # MAP_EXTERNCALL int MAP_End ( InputData *u, ParameterData *p, ContinuousData *x, ConstraintData *z, ModelData *data, OutputData *y, char *map_msg, MAP_ERROR_CODE *ierr )
    # MAP_EXTERNCALL void MAP_Input_Delete( InputData* u )
    # MAP_EXTERNCALL void MAP_Param_Delete( ParameterData* p )
    # MAP_EXTERNCALL void MAP_ContState_Delete( InputData* x )
    # MAP_EXTERNCALL void MAP_ConstrState_Delete( InputData* z )
    # MAP_EXTERNCALL void MAP_Output_Delete( InputData* y )
    # MAP_EXTERNCALL void MAP_OtherState_Delete( ModelData* data )
    def end(self):
        libexec.map_end(self.f_type_u, self.f_type_p, self.f_type_x, self.f_type_z, self.f_type_d, self.f_type_y, pointer(self.ierr), self.status)



    # Set a name for the MAP summary file. Does not need to be called. If not called, the default name is 'outlist.sum.map'
    def summary_file(self, echo_file):
        self.f_type_init.contents.summaryFileName = six.b( echo_file )
        libexec.map_set_summary_file_name(self.f_type_init, self.status, pointer(self.ierr) )



    # Calls function in fortdatamanager.c to create instance of c structs
    # MAP_EXTERNCALL InitializationData* MAP_InitInput_Create( char* map_msg, MAP_ERROR_CODE* ierr )
    def CreateInitState( self ) :
        obj = libexec.map_create_init_type( self.status, pointer(self.ierr) )
        if self.ierr.value != 0 : print(self.status.value)
        return obj

    # Calls function in fortdatamanager.c to create instance of c structs
    # MAP_EXTERNCALL void MAP_InitOutput_Delete( InputData* io )
    def CreateInitoutState( self ) :
        obj = libexec.map_create_initout_type( self.status, pointer(self.ierr) )
        if self.ierr.value != 0 : print(self.status.value)
        return obj


    # Calls function in fortdatamanager.c to create instance of c structs
    # MAP_EXTERNCALL ModelData *MAP_OtherState_Create( char *map_msg, MAP_ERROR_CODE *ierr )
    def CreateDataState( self ) :
        obj = libexec.map_create_other_type( self.status, pointer(self.ierr) )
        if self.ierr.value != 0 : print(self.status.value)
        return obj


    # Calls function in fortdatamanager.c to create instance of c structs
    # MAP_EXTERNCALL InputData* MAP_Input_Create( char* map_msg, MAP_ERROR_CODE *ierr )
    def CreateInputState( self ) :
        obj = libexec.map_create_input_type( self.status, pointer(self.ierr) )
        if self.ierr.value != 0 : print(self.status.value)
        return obj


    # Calls function in fortdatamanager.c to create instance of c structs
    # MAP_EXTERNCALL ContinuousData* MAP_ContState_Create( char* map_msg, MAP_ERROR_CODE *ierr )
    def CreateContinuousState( self ) :
        obj = libexec.map_create_continuous_type( self.status, pointer(self.ierr) )
        if self.ierr.value != 0 : print(self.status.value)
        return obj

    # Calls function in fortdatamanager.c to create instance of c structs
    # MAP_EXTERNCALL OutputData *MAP_Output_Create( char *map_msg, MAP_ERROR_CODE *ierr )
    def CreateOutputState( self ) :
        obj = libexec.map_create_output_type( self.status, pointer(self.ierr) )
        if self.ierr.value != 0 : print(self.status.value)
        return obj


    # Calls function in fortdatamanager.c to create instance of c structs
    # MAP_EXTERNCALL ConstraintData* MAP_ConstrState_Create( char* map_msg, MAP_ERROR_CODE *ierr )
    def CreateConstraintState( self ) :
        obj = libexec.map_create_constraint_type( self.status, pointer(self.ierr) )
        if self.ierr.value != 0 : print(self.status.value)
        return obj


    # Calls function in fortdatamanager.c to create instance of c structs
    # MAP_EXTERNCALL ParameterData* MAP_Param_Create( char* map_msg, MAP_ERROR_CODE *ierr )
    def CreateParameterState( self ) :
        obj = libexec.map_create_parameter_type( self.status, pointer(self.ierr) )
        if self.ierr.value != 0 : print(self.status.value)
        return obj


    def map_set_sea_depth( self, depth ):
         libexec.map_set_sea_depth( self.f_type_p, depth )

    def map_set_gravity( self, g ):
        libexec.map_set_gravity( self.f_type_p, g )

    def map_set_sea_density( self, rho ):
        libexec.map_set_sea_density( self.f_type_p, rho )

    def plot_x( self, lineNum, length ) :
        arr = [None]*length
        array = POINTER(c_double)
        array = libexec.map_plot_x_array( self.f_type_d, lineNum, length, self.status, pointer(self.ierr) )        
        if self.ierr.value != 0 :
            print(self.status.value)
            self.end( )
            libexec.map_plot_array_free( array )        
            sys.exit('MAP terminated premature.')
        arr = [array[j] for j in range(length)]        
        libexec.map_plot_array_free( array )        
        return arr 

    def plot_y( self, lineNum, length ) :
        arr = [None]*length
        array = POINTER(c_double)
        array = libexec.map_plot_y_array( self.f_type_d, lineNum, length, self.status, pointer(self.ierr) )        
        if self.ierr.value != 0 :
            print(self.status.value)
            self.end( )
            libexec.map_plot_array_free( array )        
            sys.exit('MAP terminated premature.')
        arr = [array[j] for j in range(length)]        
        libexec.map_plot_array_free( array )        
        return arr 


    def plot_z( self, lineNum, length ) :
        arr = [None]*length
        array = POINTER(c_double)
        array = libexec.map_plot_z_array( self.f_type_d, lineNum, length, self.status, pointer(self.ierr) )        
        if self.ierr.value != 0 :
            print(self.status.value)
            self.end( )
            libexec.map_plot_array_free( array )        
            sys.exit('MAP terminated premature.')
        arr = [array[j] for j in range(length)]        
        libexec.map_plot_array_free( array )        
        return arr 


    def get_fairlead_force_2d(self, index):
        """Gets the horizontal and vertical fairlead force in a 2D plane along the 
        straight-line line. Must ensure update_states() is called before accessing 
        this function. The function will not solve the forces for a new vessel position
        if it updated. , otherwise the fairlead forces are not updated with the new 
        vessel position. Called C function:
        
        MAP_EXTERNCALL void map_get_fairlead_force_2d(double* H, double* V, MAP_OtherStateType_t* other_type, int index, char* map_msg, MAP_ERROR_CODE* ierr);
    
        :param index: The line number the fairlead forces are being requested for. Zero indexed
        :returns: horizontal and vertical fairlead force [N]
    
        >>> H,V = print get_fairlead_force_2d(1)        
        """
        H_ref = c_double(-999.9)
        V_ref = c_double(-999.9)
        libexec.map_get_fairlead_force_2d( pointer(H_ref), pointer(V_ref),self.f_type_d, index, self.status, pointer(self.ierr))
        return H_ref.value, V_ref.value
    
    
    def get_fairlead_force_3d(self, index):
        """Gets the horizontal and vertical fairlead force in a 3D frame along relative 
        referene global axis. Must ensure update_states() is called before accessing 
        this function. The function will not solve the forces for a new vessel position
        if it updated. , otherwise the fairlead forces are not updated with the new 
        vessel position. Called C function:
        
        MAP_EXTERNCALL void map_get_fairlead_force_3d(double* fx, double* fy, double* fz, MAP_OtherStateType_t* other_type, int index, char* map_msg, MAP_ERROR_CODE* ierr);
    
        :param index: The line number the fairlead forces are being requested for. Zero indexed
        :returns: horizontal and vertical fairlead force [N]
    
        >>> fx,fy,fz = get_fairlead_force_3d(1)        
        """
        fx = c_double(-999.9)
        fy = c_double(-999.9)
        fz = c_double(-999.9)
        libexec.map_get_fairlead_force_3d( pointer(fx), pointer(fy), pointer(fz), self.f_type_d, index, self.status, pointer(self.ierr))
        return fx.value, fy.value, fz.value


    def get_anchor_force_2d(self, index):
        """Gets the horizontal and vertical anchor force in a 2D plane along the 
        straight-line line. Must ensure update_states() is called before accessing 
        this function. The function will not solve the forces for a new vessel position
        if it updated. , otherwise the anchor forces are not updated with the new 
        vessel position. Called C function:
        
        MAP_EXTERNCALL void map_get_anchor_force_2d(double* H, double* V, MAP_OtherStateType_t* other_type, int index, char* map_msg, MAP_ERROR_CODE* ierr);
    
        :param index: The line number the anchor forces are being requested for. Zero indexed
        :returns: horizontal and vertical anchor force [N]
    
        >>> H,V = print get_anchor_force_2d(1)        
        """
        Ha_ref = c_double(-999.9)
        Va_ref = c_double(-999.9)
        libexec.map_get_anchor_force_2d( pointer(Ha_ref), pointer(Va_ref),self.f_type_d, index, self.status, pointer(self.ierr))
        return Ha_ref.value, Va_ref.value
    
    
    def get_anchor_force_3d(self, index):
        """Gets the horizontal and vertical anchor force in a 3D frame along relative 
        referene global axis. Must ensure update_states() is called before accessing 
        this function. The function will not solve the forces for a new vessel position
        if it updated. , otherwise the anchor forces are not updated with the new 
        vessel position. Called C function:
        
        MAP_EXTERNCALL void map_get_anchor_force_3d(double* fx, double* fy, double* fz, MAP_OtherStateType_t* other_type, int index, char* map_msg, MAP_ERROR_CODE* ierr);
    
        :param index: The line number the anchor forces are being requested for. Zero indexed
        :returns: horizontal and vertical anchor force [N]
    
        >>> fx,fy,fz = get_anchor_force_3d(1)        
        """
        fxa = c_double(-999.9)
        fya = c_double(-999.9)
        fza = c_double(-999.9)
        libexec.map_get_anchor_force_3d( pointer(fxa), pointer(fya), pointer(fza), self.f_type_d, index, self.status, pointer(self.ierr))
        return fxa.value, fya.value, fza.value
    









    def funcl( self, i ) :
        self.val = libexec.map_residual_function_length(self.f_type_d, i, self.status, pointer(self.ierr))
        if self.ierr.value != 0 :
            print(self.status.value)
            self.end( )
            sys.exit('MAP terminated premature.')
        return self.val


    def funch( self, i ) :
        self.val = libexec.map_residual_function_height(self.f_type_d, i, self.status, pointer(self.ierr))
        if self.ierr.value != 0 :
            print(self.status.value)
            self.end( )
            sys.exit('MAP terminated premature.')
        return self.val


    def dxdh( self, i ) :
        self.val = libexec.map_jacobian_dxdh( self.f_type_d, i, self.status, pointer(self.ierr) )
        if self.ierr.value != 0 :
            print(self.status.value)
            self.end( )
            sys.exit('MAP terminated premature.')
        return self.val


    def dxdv( self, i ) :
        self.val = libexec.map_jacobian_dxdv( self.f_type_d, i, self.status, pointer(self.ierr) )
        if self.ierr.value != 0 :
            print(self.status.value)
            self.end( )
            sys.exit('MAP terminated premature.')
        return self.val


    def dzdh( self, i ) :
        self.val = libexec.map_jacobian_dzdh( self.f_type_d, i, self.status, pointer(self.ierr) )
        if self.ierr.value != 0 :
            print(self.status.value)
            self.end( )
            sys.exit('MAP terminated premature.')
        return self.val


    def dzdv( self, i ) :
        self.val = libexec.map_jacobian_dzdv( self.f_type_d, i, self.status, pointer(self.ierr) )
        if self.ierr.value != 0 :
            print(self.status.value)
            self.end( )
            sys.exit('MAP terminated premature.')
        return self.val

    def linear( self, epsilon ) :
        """Insert a function and its arguments in process pool.
    
        Input is inserted in queues using a round-robin fashion. Every job is
        identified by and index that is returned by function. Not all parameters
        of original multiprocessing.Pool.apply_aync are implemented so far.
    
        :param func: Function to process.
        :type func: Callable.
        :param args: Arguments for the function to process.
        :type args: Tuple.
        :returns: Assigned job id.
        :rtype: Int.
        """ 
        array = POINTER(POINTER(c_double))
        array = libexec.map_linearize_matrix( self.f_type_u, self.f_type_p, self.f_type_d, self.f_type_y, self.f_type_z, epsilon, pointer(self.ierr), self.status)        
        if self.ierr.value != 0 :
           print(self.status.value)
           self.end( )
           sys.exit('MAP terminated premature.')
        arr = [[array[j][i] for i in range(6)] for j in range(6)]
        libexec.map_free_linearize_matrix(array)        
        return arr
    

    def displace_vessel(self,x,y,z,phi,the,psi) :
        libexec.map_offset_vessel(self.f_type_d, self.f_type_u, x,y,z,phi,the,psi, self.status, pointer(self.ierr) )
        if self.ierr.value != 0 :
            print(self.status.value)
            self.end( )
            sys.exit('MAP terminated premature.')    

    def read_file( self, fileName ):
        with open(fileName) as f:
            lines = f.read().splitlines()
        self.read_list_input( lines )
                    
    def read_list_input(self, listIn):
        assert isinstance(listIn, list), 'Must input a python list of strings'
        assert len(listIn) >= 8, 'Must have at least 4 sections, 3 lines per section'
        assert type(listIn[0]) == type(''), 'List elements must be strings'
        charptr = POINTER(c_char)

        dictFlag = nodeFlag = propFlag = solvFlag = False
        listIter = iter(listIn)
        for line in listIter:
            if "LINE DICTIONARY" in line.upper():
                dictFlag = True
                nodeFlag = propFlag = solvFlag = False
                for _ in range(2): next(listIter) # Process header
                continue;
            elif "NODE PROPERTIES" in line.upper():
                nodeFlag = True
                dictFlag = propFlag = solvFlag = False
                for _ in range(2): next(listIter) # Process header
                continue;
            elif "LINE PROPERTIES" in line.upper():
                propFlag = True
                dictFlag = nodeFlag = solvFlag = False
                for _ in range(2): next(listIter) # Process header
                continue;
            elif "SOLVER OPTIONS" in line.upper():
                solvFlag = True
                dictFlag = nodeFlag = propFlag = False
                for _ in range(2): next(listIter) # Process header
                continue;

            if dictFlag:
                self.f_type_init.contents.libraryInputLine =  six.b(line+'\n\0')
                libexec.map_add_cable_library_input_text(self.f_type_init)                    

            if nodeFlag:
                self.f_type_init.contents.nodeInputLine = six.b(line+'\n\0')
                libexec.map_add_node_input_text(self.f_type_init)

            if propFlag:
                self.f_type_init.contents.elementInputLine = six.b(line+'\n\0')
                libexec.map_add_line_input_text(self.f_type_init)

            if solvFlag:
                self.f_type_init.contents.optionInputLine = six.b(line+'\n\0')
                libexec.map_add_options_input_text(self.f_type_init)            
                



import numpy as np
import numpy.testing as npt
import unittest
import wisdem.rotorse.rotor_aeropower as ra
import openmdao.api as om
import copy
import time
import os
ARCHIVE  = os.path.dirname(os.path.abspath(__file__)) + os.path.sep + 'regulation.npz'
        

class TestRotorAero(unittest.TestCase):
    def setUp(self):
        self.inputs = {}
        self.outputs = {}
        self.discrete_inputs = {}
        self.discrete_outputs = {}

        
    def testRegulationTrajectory(self):
        # Load in airfoil and blade shape inputs for NREL 5MW
        npzfile = np.load(ARCHIVE)
        self.inputs['airfoils_aoa'] = npzfile['aoa']
        self.inputs['airfoils_Re'] = npzfile['Re']
        self.inputs['airfoils_cl'] = npzfile['cl']
        self.inputs['airfoils_cd'] = npzfile['cd']
        self.inputs['airfoils_cm'] = npzfile['cm']
        self.inputs['r'] = npzfile['r']
        self.inputs['chord'] = npzfile['chord']
        self.inputs['theta'] = npzfile['theta']

        naero = self.inputs['r'].size
        n_aoa_grid = self.inputs['airfoils_aoa'].size
        n_Re_grid = self.inputs['airfoils_Re'].size
        n_pc = 22
        
        # parameters
        self.inputs['control_Vin'] = 4.
        self.inputs['control_Vout'] = 25.
        self.inputs['control_ratedPower'] = 5e6
        self.inputs['control_minOmega'] = 0.0
        self.inputs['control_maxOmega'] = 100.0
        self.inputs['control_maxTS'] = 90.
        self.inputs['control_tsr'] = 10.
        self.inputs['control_pitch'] = 0.0
        self.discrete_inputs['drivetrainType'] = 'GEARED'
        self.inputs['drivetrainEff'] = 0.95
        
        self.inputs['Rhub'] = 1.
        self.inputs['Rtip'] = 70.
        self.inputs['hub_height'] = 100.
        self.inputs['precone'] = 0.
        self.inputs['tilt'] = 0.
        self.inputs['yaw'] = 0.
        self.inputs['precurve'] = np.zeros(naero)
        self.inputs['precurveTip'] = 0.
        self.inputs['presweep'] = np.zeros(naero)
        self.inputs['presweepTip'] = 0.
        
        self.discrete_inputs['nBlades'] = 3
        self.inputs['rho'] = 1.225
        self.inputs['mu'] = 1.81206e-5
        self.inputs['shearExp'] = 0.25
        self.discrete_inputs['nSector'] = 4
        self.discrete_inputs['tiploss'] = True
        self.discrete_inputs['hubloss'] = True
        self.discrete_inputs['wakerotation'] = True
        self.discrete_inputs['usecd'] = True

        myobj = ra.RegulatedPowerCurve(naero=naero, n_aoa_grid=n_aoa_grid, n_Re_grid=n_Re_grid, n_pc=n_pc, n_pc_spline=n_pc,
                                          regulation_reg_II5=True, regulation_reg_III=True)
        myobj.naero = naero
        
        # All reg 2: no maxTS, no max rpm, no power limit
        self.inputs['control_maxOmega'] = 1e3
        self.inputs['control_maxTS'] = 1e5
        self.inputs['control_ratedPower'] = 1e16
        myobj.compute(self.inputs, self.outputs, self.discrete_inputs, self.discrete_outputs)
        V_expect0 = np.linspace(4, 25, n_pc)
        V_expect1 = V_expect0.copy()
        #V_expect1[7] = self.outputs['rated_V']
        Omega_tsr = V_expect1*10*60/70./2./np.pi
        npt.assert_equal(self.outputs['V'], V_expect1)
        npt.assert_equal(self.outputs['V_spline'], V_expect0)
        npt.assert_allclose(self.outputs['Omega'], Omega_tsr)
        npt.assert_equal(self.outputs['pitch'], np.zeros( V_expect0.shape ) )
        npt.assert_equal(self.outputs['Cp'], self.outputs['Cp_aero']*0.95)
        npt.assert_allclose(self.outputs['Cp'], self.outputs['Cp'][0])
        npt.assert_allclose(self.outputs['Cp_aero'], self.outputs['Cp_aero'][0])
        myCp = self.outputs['P']/(0.5*1.225*V_expect1**3.*np.pi*70**2)
        npt.assert_allclose(myCp, myCp[0])
        self.assertGreater(myCp[0], 0.4)
        self.assertGreater(0.5, myCp[0])
        npt.assert_allclose(myCp, self.outputs['Cp'])
        npt.assert_array_less(self.outputs['P'][:-1], self.outputs['P'][1:])
        npt.assert_array_less(self.outputs['Q'][:-1], self.outputs['Q'][1:])
        npt.assert_array_less(self.outputs['T'][:-1], self.outputs['T'][1:])
        self.assertEqual(self.outputs['rated_V'], V_expect1[-1])
        self.assertAlmostEqual(self.outputs['rated_Omega'], Omega_tsr[-1])
        self.assertEqual(self.outputs['rated_pitch'], 0.0)
        
        # Test no maxTS, max rpm, no power limit
        self.inputs['control_maxOmega'] = 15.0
        self.inputs['control_maxTS'] = 1e5
        self.inputs['control_ratedPower'] = 1e16
        myobj.compute(self.inputs, self.outputs, self.discrete_inputs, self.discrete_outputs)
        V_expect0 = np.linspace(4, 25, n_pc)
        V_expect1 = V_expect0.copy()
        #V_expect1[7] = 15.*70*2*np.pi/(10.*60.)
        Omega_tsr = V_expect1*10*60/70./2./np.pi
        Omega_expect = np.minimum(Omega_tsr, 15.0)
        npt.assert_allclose(self.outputs['V'], V_expect1)
        npt.assert_equal(self.outputs['V_spline'], V_expect0)
        npt.assert_allclose(self.outputs['Omega'], Omega_expect)
        npt.assert_equal(self.outputs['pitch'][:7], 0.0 )
        npt.assert_array_less(0.0, np.abs(self.outputs['pitch'][7:]))
        npt.assert_equal(self.outputs['Cp'], self.outputs['Cp_aero']*0.95)
        npt.assert_array_less(self.outputs['P'][:-1], self.outputs['P'][1:])
        npt.assert_array_less(self.outputs['Q'][:-1], self.outputs['Q'][1:])
        npt.assert_array_less(self.outputs['T'][:-1], self.outputs['T'][1:])
        self.assertAlmostEqual(self.outputs['rated_V'], V_expect1[-1], 3)
        self.assertAlmostEqual(self.outputs['rated_Omega'], 15.0)
        self.assertGreater(self.outputs['rated_pitch'], 0.0)
        myCp = self.outputs['P']/(0.5*1.225*V_expect1**3.*np.pi*70**2)
        npt.assert_allclose(myCp[:7], myCp[0])
        npt.assert_allclose(myCp[:7], self.outputs['Cp'][:7])

        # Test maxTS, no max rpm, no power limit
        self.inputs['control_maxOmega'] = 1e3
        self.inputs['control_maxTS'] = 105.0
        self.inputs['control_ratedPower'] = 1e16
        myobj.compute(self.inputs, self.outputs, self.discrete_inputs, self.discrete_outputs)
        V_expect0 = np.linspace(4, 25, n_pc)
        V_expect1 = V_expect0.copy()
        #V_expect1[7] = 105./10.
        Omega_tsr = V_expect1*10*60/70./2./np.pi
        Omega_expect = np.minimum(Omega_tsr, 105./70./2/np.pi*60)
        npt.assert_allclose(self.outputs['V'], V_expect1)
        npt.assert_equal(self.outputs['V_spline'], V_expect0)
        npt.assert_allclose(self.outputs['Omega'], Omega_expect)
        npt.assert_equal(self.outputs['pitch'][:7], 0.0 )
        npt.assert_array_less(0.0, np.abs(self.outputs['pitch'][7:]))
        npt.assert_equal(self.outputs['Cp'], self.outputs['Cp_aero']*0.95) 
        npt.assert_array_less(self.outputs['P'][:-1], self.outputs['P'][1:])
        npt.assert_array_less(self.outputs['Q'][:-1], self.outputs['Q'][1:])
        npt.assert_array_less(self.outputs['T'][:-1], self.outputs['T'][1:])
        self.assertEqual(self.outputs['rated_V'], V_expect1[-1])
        self.assertAlmostEqual(self.outputs['rated_Omega'], Omega_expect[-1])
        self.assertGreater(self.outputs['rated_pitch'], 0.0)
        myCp = self.outputs['P']/(0.5*1.225*V_expect1**3.*np.pi*70**2)
        npt.assert_allclose(myCp[:7], myCp[0])
        npt.assert_allclose(myCp[:7], self.outputs['Cp'][:7])

        # Test no maxTS, no max rpm, power limit
        self.inputs['control_maxOmega'] = 1e3
        self.inputs['control_maxTS'] = 1e4
        self.inputs['control_ratedPower'] = 5e6
        myobj.compute(self.inputs, self.outputs, self.discrete_inputs, self.discrete_outputs)
        V_expect0 = np.linspace(4, 25, n_pc)
        V_expect1 = V_expect0.copy()
        V_expect1[7] = self.outputs['rated_V']
        Omega_tsr = V_expect1*10*60/70./2./np.pi
        Omega_expect = np.minimum(Omega_tsr, self.outputs['rated_Omega'])
        npt.assert_allclose(self.outputs['V'], V_expect1)
        npt.assert_equal(self.outputs['V_spline'], V_expect0)
        npt.assert_allclose(self.outputs['Omega'], Omega_expect)
        npt.assert_equal(self.outputs['pitch'][:7], 0.0 )
        npt.assert_array_less(0.0, np.abs(self.outputs['pitch'][8:]))
        npt.assert_equal(self.outputs['Cp'], self.outputs['Cp_aero']*0.95)
        npt.assert_array_less(self.outputs['P'][:7], self.outputs['P'][1:8])
        npt.assert_allclose(self.outputs['P'][7:], 5e6, rtol=1e-4, atol=0)
        #npt.assert_array_less(self.outputs['Q'], self.outputs['Q'][1:])
        npt.assert_array_less(self.outputs['T'], self.outputs['T'][7]+1e-1)
        #self.assertEqual(self.outputs['rated_V'], V_expect1[-1])
        self.assertAlmostEqual(self.outputs['rated_Omega'], Omega_expect[-1])
        self.assertEqual(self.outputs['rated_pitch'], 0.0)
        myCp = self.outputs['P']/(0.5*1.225*V_expect1**3.*np.pi*70**2)
        npt.assert_allclose(myCp[:7], myCp[0])
        npt.assert_allclose(myCp[:7], self.outputs['Cp'][:7])
        
        # Test min & max rpm, no power limit
        self.inputs['control_minOmega'] = 7.0
        self.inputs['control_maxOmega'] = 15.0
        self.inputs['control_maxTS'] = 1e5
        self.inputs['control_ratedPower'] = 1e16
        myobj.compute(self.inputs, self.outputs, self.discrete_inputs, self.discrete_outputs)
        V_expect0 = np.linspace(4, 25, n_pc)
        V_expect1 = V_expect0.copy()
        #V_expect1[7] = 15.*70*2*np.pi/(10.*60.)
        Omega_tsr = V_expect1*10*60/70./2./np.pi
        Omega_expect = np.maximum( np.minimum(Omega_tsr, 15.0), 7.0)
        npt.assert_allclose(self.outputs['V'], V_expect1)
        npt.assert_equal(self.outputs['V_spline'], V_expect0)
        npt.assert_allclose(self.outputs['Omega'], Omega_expect)
        npt.assert_array_less(0.0, np.abs(self.outputs['pitch'][Omega_expect != Omega_tsr]) )
        npt.assert_equal(self.outputs['Cp'], self.outputs['Cp_aero']*0.95) 
        npt.assert_array_less(self.outputs['P'][:-1], self.outputs['P'][1:])
        npt.assert_array_less(self.outputs['Q'][:-1], self.outputs['Q'][1:])
        npt.assert_array_less(self.outputs['T'][:-1], self.outputs['T'][1:])
        self.assertEqual(self.outputs['rated_V'], V_expect1[-1])
        self.assertAlmostEqual(self.outputs['rated_Omega'], 15.0)
        self.assertGreater(self.outputs['rated_pitch'], 0.0)
        myCp = self.outputs['P']/(0.5*1.225*V_expect1**3.*np.pi*70**2)
        npt.assert_allclose(myCp[Omega_expect == Omega_tsr], myCp[6])
        npt.assert_allclose(myCp[Omega_expect == Omega_tsr], self.outputs['Cp'][Omega_expect == Omega_tsr])
        
        # Test fixed pitch
        self.inputs['control_minOmega'] = 0.0
        self.inputs['control_maxOmega'] = 15.0
        self.inputs['control_maxTS'] = 1e5
        self.inputs['control_ratedPower'] = 1e16
        self.inputs['control_pitch'] = 5.0
        myobj.compute(self.inputs, self.outputs, self.discrete_inputs, self.discrete_outputs)
        V_expect0 = np.linspace(4, 25, n_pc)
        V_expect1 = V_expect0.copy()
        #V_expect1[7] = 15.*70*2*np.pi/(10.*60.)
        Omega_tsr = V_expect1*10*60/70./2./np.pi
        Omega_expect = np.minimum(Omega_tsr, 15.0)
        npt.assert_allclose(self.outputs['V'], V_expect1)
        npt.assert_equal(self.outputs['V_spline'], V_expect0)
        npt.assert_allclose(self.outputs['Omega'], Omega_expect)
        npt.assert_equal(self.outputs['pitch'][:7], 5.0 )
        npt.assert_array_less(0.0, np.abs(self.outputs['pitch'][7:]))
        npt.assert_equal(self.outputs['Cp'], self.outputs['Cp_aero']*0.95)
        npt.assert_array_less(self.outputs['P'][:-1], self.outputs['P'][1:])
        npt.assert_array_less(self.outputs['Q'][:-1], self.outputs['Q'][1:])
        npt.assert_array_less(self.outputs['T'][:-1], self.outputs['T'][1:])
        self.assertAlmostEqual(self.outputs['rated_V'], V_expect1[-1], 3)
        self.assertAlmostEqual(self.outputs['rated_Omega'], 15.0)
        self.assertGreater(self.outputs['rated_pitch'], 5.0)
        myCp = self.outputs['P']/(0.5*1.225*V_expect1**3.*np.pi*70**2)
        npt.assert_allclose(myCp[:7], myCp[0])
        npt.assert_allclose(myCp[:7], self.outputs['Cp'][:7])

        
        
    def testRegulationTrajectoryNoRegion3(self):
        # Load in airfoil and blade shape inputs for NREL 5MW
        npzfile = np.load(ARCHIVE)
        self.inputs['airfoils_aoa'] = npzfile['aoa']
        self.inputs['airfoils_Re'] = npzfile['Re']
        self.inputs['airfoils_cl'] = npzfile['cl']
        self.inputs['airfoils_cd'] = npzfile['cd']
        self.inputs['airfoils_cm'] = npzfile['cm']
        self.inputs['r'] = npzfile['r']
        self.inputs['chord'] = npzfile['chord']
        self.inputs['theta'] = npzfile['theta']

        naero = self.inputs['r'].size
        n_aoa_grid = self.inputs['airfoils_aoa'].size
        n_Re_grid = self.inputs['airfoils_Re'].size
        n_pc = 22
        
        # parameters
        self.inputs['control_Vin'] = 4.
        self.inputs['control_Vout'] = 25.
        self.inputs['control_ratedPower'] = 5e6
        self.inputs['control_minOmega'] = 0.0
        self.inputs['control_maxOmega'] = 100.0
        self.inputs['control_maxTS'] = 90.
        self.inputs['control_tsr'] = 10.
        self.inputs['control_pitch'] = 0.0
        self.discrete_inputs['drivetrainType'] = 'GEARED'
        self.inputs['drivetrainEff'] = 0.95
        
        self.inputs['Rhub'] = 1.
        self.inputs['Rtip'] = 70.
        self.inputs['hub_height'] = 100.
        self.inputs['precone'] = 0.
        self.inputs['tilt'] = 0.
        self.inputs['yaw'] = 0.
        self.inputs['precurve'] = np.zeros(naero)
        self.inputs['precurveTip'] = 0.
        self.inputs['presweep'] = np.zeros(naero)
        self.inputs['presweepTip'] = 0.
        
        self.discrete_inputs['nBlades'] = 3
        self.inputs['rho'] = 1.225
        self.inputs['mu'] = 1.81206e-5
        self.inputs['shearExp'] = 0.25
        self.discrete_inputs['nSector'] = 4
        self.discrete_inputs['tiploss'] = True
        self.discrete_inputs['hubloss'] = True
        self.discrete_inputs['wakerotation'] = True
        self.discrete_inputs['usecd'] = True

        myobj = ra.RegulatedPowerCurve(naero=naero, n_aoa_grid=n_aoa_grid, n_Re_grid=n_Re_grid, n_pc=n_pc, n_pc_spline=n_pc,
                                          regulation_reg_II5=True, regulation_reg_III=False)
        myobj.naero = naero
        
        # All reg 2: no maxTS, no max rpm, no power limit
        self.inputs['control_maxOmega'] = 1e3
        self.inputs['control_maxTS'] = 1e5
        self.inputs['control_ratedPower'] = 1e16
        myobj.compute(self.inputs, self.outputs, self.discrete_inputs, self.discrete_outputs)
        V_expect0 = np.linspace(4, 25, n_pc)
        V_expect1 = V_expect0.copy()
        #V_expect1[7] = self.outputs['rated_V']
        Omega_tsr = V_expect1*10*60/70./2./np.pi
        npt.assert_equal(self.outputs['V'], V_expect1)
        npt.assert_equal(self.outputs['V_spline'], V_expect0)
        npt.assert_allclose(self.outputs['Omega'], Omega_tsr)
        npt.assert_equal(self.outputs['pitch'], np.zeros( V_expect0.shape ) )
        npt.assert_equal(self.outputs['Cp'], self.outputs['Cp_aero']*0.95)
        npt.assert_allclose(self.outputs['Cp'], self.outputs['Cp'][0])
        npt.assert_allclose(self.outputs['Cp_aero'], self.outputs['Cp_aero'][0])
        myCp = self.outputs['P']/(0.5*1.225*V_expect1**3.*np.pi*70**2)
        npt.assert_allclose(myCp, myCp[0])
        self.assertGreater(myCp[0], 0.4)
        self.assertGreater(0.5, myCp[0])
        npt.assert_allclose(myCp, self.outputs['Cp'])
        npt.assert_array_less(self.outputs['P'][:-1], self.outputs['P'][1:])
        npt.assert_array_less(self.outputs['Q'][:-1], self.outputs['Q'][1:])
        npt.assert_array_less(self.outputs['T'][:-1], self.outputs['T'][1:])
        self.assertEqual(self.outputs['rated_V'], V_expect1[-1])
        self.assertAlmostEqual(self.outputs['rated_Omega'], Omega_tsr[-1])
        self.assertEqual(self.outputs['rated_pitch'], 0.0)
        
        # Test no maxTS, no max rpm, power limit
        self.inputs['control_maxOmega'] = 1e3
        self.inputs['control_maxTS'] = 1e4
        self.inputs['control_ratedPower'] = 5e6
        myobj.compute(self.inputs, self.outputs, self.discrete_inputs, self.discrete_outputs)
        V_expect0 = np.linspace(4, 25, n_pc)
        V_expect1 = V_expect0.copy()
        V_expect1[7] = self.outputs['rated_V']
        Omega_tsr = V_expect1*10*60/70./2./np.pi
        Omega_expect = np.minimum(Omega_tsr, self.outputs['rated_Omega'])
        npt.assert_allclose(self.outputs['V'], V_expect1)
        npt.assert_equal(self.outputs['V_spline'], V_expect0)
        npt.assert_allclose(self.outputs['Omega'], Omega_expect)
        npt.assert_equal(self.outputs['pitch'], 0.0 )
        #npt.assert_array_less(0.0, np.abs(self.outputs['pitch'][8:]))
        npt.assert_allclose(self.outputs['Cp'][:8], self.outputs['Cp_aero'][:8]*0.95)
        npt.assert_array_less(self.outputs['P'][:7], self.outputs['P'][1:8])
        npt.assert_allclose(self.outputs['P'][7:], 5e6, rtol=1e-6, atol=0)
        #npt.assert_equal(self.outputs['Q'][8:], self.outputs['Q'][7])
        npt.assert_equal(self.outputs['T'][8:], 0.0)
        npt.assert_array_less(self.outputs['T'], self.outputs['T'][7]+1e-1)
        #self.assertEqual(self.outputs['rated_V'], V_expect1[-1])
        self.assertAlmostEqual(self.outputs['rated_Omega'], Omega_expect[-1])
        self.assertEqual(self.outputs['rated_pitch'], 0.0)
        myCp = self.outputs['P']/(0.5*1.225*V_expect1**3.*np.pi*70**2)
        npt.assert_allclose(myCp[:7], myCp[0])
        npt.assert_allclose(myCp[:7], self.outputs['Cp'][:7])
        
def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestRotorAero))
    return suite

if __name__ == '__main__':
    unittest.TextTestRunner().run(suite())



"""
csmPPI.py

Created by George Scott on 2012-08-01.
Modified by Katherine Dykes 2012.
Copyright (c) NREL. All rights reserved.
"""
import wisdem.nrelcsm as nrelcsm
import os
import sys
import re
 
class Escalator:
    ''' 
        cost escalator class - computes weighted sum of different PPI_tables 
    '''
    
    def __init__(self,x):
        ''' x is a list '''
        self.name = x[0]  # name
        self.tbls = x[1]  # list of table names
        self.wts  = x[2]  # list of weights in pct (0-100.0)
        
        sum = 0.0
        for i in range(len(self.wts)):
            self.wts[i] *= 0.01
            sum += self.wts[i]
        if (abs(sum-1.0) > 0.0001):
            print('Weights add up to {0:.5f}'.format(sum))
            
    def compute(self,ppitbls,sy,sm,ey,em):
        ''' 
            returns cost escalator between start_yr/start_mon and end_yr/end_mon 
            ppitbls is a dictionary of PPITbl objects, indexed by NAICS code
        '''
        sum = 0.0
        for i in range(len(self.tbls)):
            key = self.tbls[i].strip()
            '''if (key == 'GDP'):
                print('Skipping GDP')
                continue'''
            if (key not in ppitbls):
                print ('No PPI table %s'.format(key))
                continue           
            if (key == "326150P") and (self.name == "Advanced Blade material costs       "):
            		ce = ppitbls[key].getEsc(2002,sm,ey,em)
            else:
            		ce = ppitbls[key].getEsc(sy,sm,ey,em)
            sum += ce * self.wts[i]
        return sum

#------------------------------------------------

class PPITbl:
    ''' a PPITbl object represents a cost table '''
    
    def __init__(self,code="",name="PPITbl"):
        self.cost = [] # empty list
        self.years = []
        self.name = name
        self.code = code
        pass
        
    def add_row(self,year,cost_array):
        self.cost.append(cost_array)
        self.years.append(year) 
        return 1
        
    def getEsc(self,start_yr,start_mon,end_yr,end_mon,printFlag=0):
        ''' return cost escalator between two dates (mon==13 is annual value) '''
        start_row = start_yr-self.years[0]
        end_row   = end_yr-self.years[0]
        
        if (start_row < 0):
            print('\n*** Year start_yr ${0:.2f} before table start {1}\n'.format(start_yr,self.years[0]))
            return None
        if (end_row >= len(self.cost)):
            print('\n*** Year end_yr ${0:.2f} after table end {1}\n'.format(end_yr,self.years[-1]))
            return None            
        if (len(self.cost[start_row]) < start_mon):
            raise IndexError("Start_mon out of range")
        if (len(self.cost[int(end_row)]) < int(end_mon)):
            print('\n*** EM {0:2d} > LER {0:2d} in table {}'.format(end_mon, len(self.cost[end_row]), self.code))
            raise IndexError("End_mon out of range")
            
        try:
            '''print('SR {0:2d} ER {0:2d}'.format(start_row,end_row))
            print('LSR {0:2d} LER {0:2d}'.format(len(self.cost[start_row]),len(self.cost[end_row])))'''
            cost_start = self.cost[start_row][start_mon-1]
            cost_end   = self.cost[int(end_row)][int(end_mon)-1]
        except:
            print('Index out of range for table {} {}'.format(self.code, self.name))
            return None
        esc = cost_end / cost_start
        return esc

#--------------------------------------------------------------------------------------

class PPI:
    def __init__(self,ref_yr,ref_mon,curr_yr,curr_mon,debug=0):
        '''
        Initialize the PPI class for calculation of PPI indices given a referene year/month and current year/month.
        
        Parameters
        ----------
        ref_yr : int
          reference PPI year
        ref_mon : int
          reference PPI month
        curr_yr : int
          current PPI year
        curr_mon : int
          current PPI month   
        '''     
        
        #self.escData = [None] * 37
        self.escData = {}  # try a dictionary
        self.tblfile = os.path.join('static','PPI_Tables.txt')   #TODO: temporary solution - should update so it can locate it from dictionary etc
        self.ppitbls = {} # dictionary of PPITbl objects 
        self.yrs_gdp = []
        self.ppi_gdp = []
        
        self.ref_yr   = ref_yr
        self.ref_mon  = ref_mon
        self.curr_yr  = curr_yr
        self.curr_mon = curr_mon
        self.debug = debug

        fullfile = os.path.join(os.path.dirname(os.path.abspath(__file__)), self.tblfile)
        try:
            infile = open(fullfile, 'r') #infile = open(self.tblfile)
        except:
            sys.stdout.write ("Error opening or reading %s\n" % fullfile)
            quit()
        else:
            if (self.debug > 0):
                sys.stdout.write ("Opened %s\n" % self.tblfile)
            
        itable = -1
        found_tables = False
        found_GDP = False
        iCode   = {}
        code = ''
    
        for line in infile:
            words = line.split("\t")
            if not words:
                continue
                
            if (words[0].startswith("Gross Domestic Product")):
                found_GDP = True
                continue
            if (found_GDP and words[0].startswith("Year")):
                y = []
                for i in range(1,len(words)):
                    if (words[i].startswith("20")):
                        y.append(int(words[i]))
                self.yrs_gdp = y
                continue
            if (found_GDP and words[0].startswith("Absolute Value")):
                g = []
                num_re = re.compile(r"[\d\.]+")
                for i in range(1,len(words)):
                    if (num_re.search(words[i])):
                        g.append(float(words[i]))
                self.ppi_gdp = g
                
                code = 'GDP'
                self.ppitbls[code] = PPITbl(code=code, name="Gross Domestic Product")  # add a new element to self.ppitbls
                for i in range(len(y)):
                    rvals = []
                    for mon in range(13):
                        rvals.append(g[i]) # fill all monthly values with annual value
                    self.ppitbls[code].add_row(y[i], rvals)
                ippi = len(self.ppitbls)-1
                iCode[code] = ippi  # index of tables by code
                if (self.debug > 0): 
                    print('Created {0:2d} {} {}'.format(ippi, code, self.ppitbls[code].name))
                continue
                
            if (words[0].startswith("NAICS")):
                found_tables = True
                itable += 1
                iyr = 0
                
                words[2] = words[2].replace(r'"', '')  # strip quotes from name
                code = words[1]
                self.ppitbls[code] = PPITbl(code=code, name=words[2])  # add a new element to self.ppitbls
                ippi = len(self.ppitbls)-1
                if (self.debug > 0): 
                    print('Created {0:2d} {} {}'.format(ippi, code, self.ppitbls[code].name))
                iCode[code] = ippi  # index of tables by code 
                
            if (found_tables and words[0].startswith("20")): # a year number
                rvals = []
                i = 1
                while (re.match(r"\d+\.",words[i])):
                    rvals.append(float(words[i]))
                    i += 1
                self.ppitbls[code].add_row(int(words[0]), rvals)
                iyr += 1
        
        self.escData['IPPI_BLD'] = Escalator( ['Baseline Blade material costs       ',   ['3272123', '3255204', '332722489', '326150P'], [ 60.00,  23.00,  8.00,   9.00 ]  ] )
        self.escData['IPPI_BLA'] = Escalator( ['Advanced Blade material costs       ',   ['3272123', '3255204', '332722489', '326150P'], [ 61.00,  27.00,  3.00,   9.00 ]  ] )
        self.escData['IPPI_BLL'] = Escalator( ['Blade Labor costs                   ',   ['GDP    '                                   ], [ 100.00                       ]  ] )
        self.escData['IPPI_HUB'] = Escalator( ['Hub                                 ',   ['3315113',                                  ], [ 100.00                       ]  ] )
        self.escData['IPPI_PMB'] = Escalator( ['Pitch Mechanisms/Bearings           ',   ['332991P', '3353123', '333612P  ', '334513' ], [ 50.00,  20.00, 20.00,  10.00 ]  ] )
        self.escData['IPPI_LSS'] = Escalator( ['Low speed shaft                     ',   ['3315131'                                   ], [ 100.00                       ]  ] )
        self.escData['IPPI_BRN'] = Escalator( ['Bearings                            ',   ['332991P'                                   ], [ 100.00                       ]  ] )
        self.escData['IPPI_GRB'] = Escalator( ['Gearbox                             ',   ['333612P'                                   ], [ 100.00                       ]  ] )
        self.escData['IPPI_BRK'] = Escalator( ['Mech brake, HS cpling etc           ',   ['3363401'                                   ], [ 100.00                       ]  ] )
        self.escData['IPPI_GEN'] = Escalator( ['Generator                           ',   ['335312P'                                   ], [ 100.00                       ]  ] )
        self.escData['IPPI_VSE'] = Escalator( ['Variable spd electronics            ',   ['335314P'                                   ], [ 100.00                       ]  ] )
        self.escData['IPPI_YAW'] = Escalator( ['Yaw drive & bearing                 ',   ['3353123', '332991P',                       ], [ 50.00, 50.00                 ]  ] )
        self.escData['IPPI_MFM'] = Escalator( ['Main frame                          ',   ['3315113'                                   ], [ 100.00                       ]  ] )
        self.escData['IPPI_ELC'] = Escalator( ['Electrical connections              ',   ['335313P', '3359291', 'GDP      ',          ], [ 25.00, 60.00, 15.00          ]  ] )
        self.escData['IPPI_HYD'] = Escalator( ['Hydraulic system                    ',   ['3339954'                                   ], [ 100.00                       ]  ] )
        self.escData['IPPI_NAC'] = Escalator( ['Nacelle                             ',   ['3272123', '3255204', 'GDP      ',          ], [ 55.00, 30.00, 15.00          ]  ] )
        self.escData['IPPI_CTL'] = Escalator( ['Control, safety system              ',   ['334513 '                                   ], [ 100.00                       ]  ] )
        self.escData['IPPI_MAR'] = Escalator( ['Marinization                        ',   ['GDP    '                                   ], [ 100.00                       ]  ] )
        self.escData['IPPI_WAR'] = Escalator( ['Offshore Warranty Premium           ',   ['GDP    '                                   ], [ 100.00                       ]  ] )
        self.escData['IPPI_TWR'] = Escalator( ['Tower                               ',   ['331221 '                                   ], [ 100.00                       ]  ] )
        self.escData['IPPI_MPF'] = Escalator( ['Monopole Foundations                ',   ['BHVY   '                                   ], [ 100.00                       ]  ] )
        self.escData['IPPI_TPT'] = Escalator( ['Transportation On/Offshore          ',   ['4841212'                                   ], [ 100.00                       ]  ] )
        self.escData['IPPI_STP'] = Escalator( ['Off Shore Site Prep                 ',   ['BHVY   '                                   ], [ 100.00                       ]  ] )
        self.escData['IPPI_LAI'] = Escalator( ['Land Based Assembly & installation  ',   ['BHVY   '                                   ], [ 100.00                       ]  ] )
        self.escData['IPPI_OAI'] = Escalator( ['Offshore Assembly & installation    ',   ['BHVY   '                                   ], [ 100.00                       ]  ] )
        self.escData['IPPI_LEL'] = Escalator( ['Land Based Elect                    ',   ['3353119', '335313P', '3359291  ', 'GDP'    ], [ 40.00, 15.00, 35.00, 10.00   ]  ] )
        self.escData['IPPI_OEL'] = Escalator( ['Offshore Elect                      ',   ['3353119', '335313P', '3359291  ', 'GDP'    ], [  5.00,  5.00, 70.00, 20.00   ]  ] )
        self.escData['IPPI_LPM'] = Escalator( ['Permits, engineering (Land Based)   ',   ['GDP    '                                   ], [ 100.00                       ]  ] )
        self.escData['IPPI_OPM'] = Escalator( ['Permits, engineering (Offshore)     ',   ['GDP    '                                   ], [ 100.00                       ]  ] )
        self.escData['IPPI_LLR'] = Escalator( ['Land Based Levelized Replacement    ',   ['GDP    '                                   ], [ 100.00                       ]  ] )
        self.escData['IPPI_OLR'] = Escalator( ['Offshore Levelized Replacement      ',   ['GDP    '                                   ], [ 100.00                       ]  ] )
        self.escData['IPPI_LOM'] = Escalator( ['O&M Land Based                      ',   ['GDP    '                                   ], [ 100.00                       ]  ] )
        self.escData['IPPI_OOM'] = Escalator( ['O&M Offshore                        ',   ['GDP    '                                   ], [ 100.00                       ]  ] )
        self.escData['IPPI_LSE'] = Escalator( ['Land Based & Offshore Lease Cost    ',   ['GDP    '                                   ], [ 100.00                       ]  ] )
        self.escData['IPPI_FND'] = Escalator( ['Foundations                         ',   ['BHVY   '                                   ], [ 100.00                       ]  ] )
        self.escData['IPPI_RDC'] = Escalator( ['Road & Civil Work                   ',   ['BHWY   '                                   ], [ 100.00                       ]  ] )
        self.escData['IPPI_PAE'] = Escalator( ['Personnel Access Equipment          ',   ['GDP    '                                   ], [ 100.00                       ]  ] )

    def compute(self,escCode,debug=0):
        """
        Returns the cost escalator for escData object 'escCode', using reference and current yr/mon values.
        
        Parameters
        ----------
        escCode : object
          codes for different PPI data   
        
        Returns
        -------
        esc : float
          cost escalator for a particular PPI from reference year/month to current year month
        """ 
        # returns the cost escalator for escData object 'escCode', using reference and current yr/mon values
        
        if (escCode not in self.escData):
            print('Warning - no such code in PPI {0:.2f}'.format(escCode))
            #return 0
        esc = self.escData[escCode].compute(self.ppitbls,self.ref_yr,self.ref_mon,self.curr_yr,self.curr_mon)
        if (debug > 0):
            print('Escalator {} from {}{:02} to {}{:02} = {:.4}'.format(escCode,self.ref_yr,self.ref_mon,self.curr_yr,self.curr_mon,esc))
        return esc
        
#--------------------------------------------------------------------------------------

def example():
    
    # test cases
    
    sy = 2002; sm = 9;
    ey = 2010; em = 13;
    ppi = PPI(sy,sm,ey,em)
         
    for i in list(ppi.escData.keys()):
        print(ppi.escData[i].name)    
        # Use ppi.compute(code) to compute ppi for default interval
        print('{0:2d} {:.4} '.format(i, ppi.compute(i)))     
        print(' ' '')

if __name__ == "__main__":

    example()                                                                                           



#-------------------------------------------------------------------------------
# Name:        Tube.py
# Purpose: This module contains the tube class, which calculates structural
#          properties of a hollow, uniform, cylindrical beam
#
# Author:      rdamiani
#
# Created:     04/11/2013
# Copyright:   (c) rdamiani 2013
# Licence:     <your licence>
#-------------------------------------------------------------------------------
from __future__ import print_function
import math
import numpy as np
from .Material import Material
from openmdao.api import ExplicitComponent, Group, Problem
from wisdem.commonse.utilities import nodal2sectional
pi = math.pi

def main():
    print('Testing Tube')
    a=Tube(np.array([5.,6.]),np.array([0.10,0.10]),np.array([10.,10]),mat=Material(matname='heavySteel'))
    print('Area and PSEUDOMass of tube a', a.Area, a.pseudomass)
    b=Tube(3.,0.10,10.,mat=Material(matname='heavySteel'))
    print('Area and PSEUDOMass of tube b', b.Area, b.pseudomass)



class CylindricalShellProperties(ExplicitComponent):
    def initialize(self):
        self.options.declare('nFull')
        
    def setup(self):
        nFull = self.options['nFull']

        self.add_input('d', np.zeros(nFull), units='m', desc='tower diameter at corresponding locations')
        self.add_input('t', np.zeros(nFull-1), units='m', desc='shell thickness at corresponding locations')

        self.add_output('Az', np.zeros(nFull-1), units='m**2', desc='cross-sectional area')
        self.add_output('Asx', np.zeros(nFull-1), units='m**2', desc='x shear area')
        self.add_output('Asy', np.zeros(nFull-1), units='m**2', desc='y shear area')
        self.add_output('Jz', np.zeros(nFull-1), units='m**4', desc='polar moment of inertia')
        self.add_output('Ixx', np.zeros(nFull-1), units='m**4', desc='area moment of inertia about x-axis')
        self.add_output('Iyy', np.zeros(nFull-1), units='m**4', desc='area moment of inertia about y-axis')

        # Derivatives
        self.declare_partials('*', '*', method='fd', form='central', step=1e-6)


    def compute(self, inputs, outputs):
        d,_ = nodal2sectional(inputs['d'])
        tube = Tube(d,inputs['t'])

        outputs['Az'] = tube.Area
        outputs['Asx'] = tube.Asx
        outputs['Asy'] = tube.Asy
        outputs['Jz'] = tube.J0
        outputs['Ixx'] = tube.Jxx
        outputs['Iyy'] = tube.Jyy

##        ro = self.d/2.0 + self.t/2.0
##        ri = self.d/2.0 - self.t/2.0
##        self.Az = math.pi * (ro**2 - ri**2)
##        self.Asx = self.Az / (0.54414 + 2.97294*(ri/ro) - 1.51899*(ri/ro)**2)
##        self.Asy = self.Az / (0.54414 + 2.97294*(ri/ro) - 1.51899*(ri/ro)**2)
##        self.Jz = math.pi/2.0 * (ro**4 - ri**4)
##        self.Ixx = self.Jz/2.0
##        self.Iyy = self.Jz/2.0



class Tube:
    """The Tube Class contains functions to calculate properties of tubular circular cross-sections
    for structural analyses. It assumes 1 material specification, even though I may pass
    more than 1 element (D's,L's,Kbuck's, and t's) to the class."""
    def __init__(self, D, t, Lgth=np.NaN, Kbuck=1., mat=Material(name='ASTM992 steel')):
        self.D=D
        self.t=t
        self.L=Lgth*np.ones(np.size(D)) #this makes sure we exapnd Lght if D,t, arrays
        self.Kbuck=Kbuck*np.ones(np.size(D)) #this makes sure we exapnd Kbuck if D,t, arrays
        self.mat=mat
        if np.size(D)>1 and type(mat) != np.ndarray: #in this case I need to make sure we have a list of materials
            import copy
            self.mat=np.array([copy.copy(mat) for i in range(np.size(D))])

    @property
    def Area(self): #Cross sectional area of tube
        return (self.D**2-(self.D-2*self.t)**2)* pi/4

    @property
    def derivArea(self):
        return {'D': 2* pi/4 * (self.D**2-(self.D-2*self.t)) * (2*self.D -1),
                't': 0}

    @property
    def Amid(self): #mid-thickness inscribed area of tube (thin wall torsion calculation)
        return (self.D-self.t)**2* pi/4

    @property
    def Jxx(self): #2nd area moment of inertia w.r.t. x-x axis (Jxx=Jyy for tube)
        return (self.D**4-(self.D-2*self.t)**4)* pi/64

    @property
    def Jyy(self): #2nd area moment of inertia w.r.t. x-x axis (Jxx=Jyy for tube)
        return self.Jxx

    @property
    def J0(self):  #polar moment of inertia w.r.t. z-z axis (torsional)
        return (2.0 * self.Jxx)

    @property
    def Asy(self): #Shear Area for tubular cross-section
        Ri=self.D/2-self.t
        Ro=self.D/2
        return self.Area / ( 1.124235 + 0.055610*(Ri/Ro) + 1.097134*(Ri/Ro)**2 - 0.630057*(Ri/Ro)**3 )

    @property
    def Asx(self): #Shear Area for tubular cross-section
        return self.Asy

    @property
    def BdgMxx(self):  #Bending modulus for tubular cross-section
        return self.Jxx / (self.D/2)

    @property
    def BdgMyy(self):  #Bending modulus for tubular cross-section =BdgMxx
        return self.Jyy / (self.D/2)

    @property
    def TorsConst(self):  #Torsion shear constant for tubular cross-section
        return self.J0 / (self.D/2)

    @property
    def S(self):  #Bending modulus for tubular cross-section
        return self.BdgMxx

    @property
    def C(self):  #Torsion shear constant for tubular cross-section
        return self.TorsConst

    @property
    def Rgyr(self): #Radius of Gyration for circular tube
        return np.sqrt(self.Jxx/self.Area)

    @property
    def Klr(self): #Klr buckling parameter
        return self.Kbuck*self.L/self.Rgyr

    @property
    def pseudomass(self): #This uses the buckling length as actual length, !!!CAUTION!!!!!!
        pseudomass =0. #initialize

        if np.size(self.D)>1:
            for ii,mat in enumerate(self.mat):
                    pseudomass +=self.Area*self.L*mat.rho
        else:
            pseudomass =(self.Area*self.L*self.mat.rho)
        return pseudomass


if __name__ == '__main__':
    main()



import unittest

from wisdem.test.test_pyframe3dd import test_frame

def suite():
    suite = unittest.TestSuite( (test_frame.suite(),
    ) )
    return suite


if __name__ == '__main__':
    unittest.TextTestRunner().run(suite())
        



"""scig.py
Created by Latha Sethuraman, Katherine Dykes.
Copyright (c) NREL. All rights reserved."""

from openmdao.api import Group, Problem, ExplicitComponent,ExecComp,IndepVarComp,ScipyOptimizeDriver
import numpy as np
from math import pi, cos, sqrt, radians, sin, exp, log10, log, tan, atan


class SCIG(ExplicitComponent):
    
    """ Estimates overall mass dimensions and Efficiency of Squirrel cage Induction generator. """
    
    def setup(self):
        
        # SCIG generator design inputs
        self.add_input('r_s', val=0.0, units ='m', desc='airgap radius r_s')
        self.add_input('l_s', val=0.0, units ='m', desc='Stator core length l_s')
        self.add_input('h_s', val=0.0, units ='m', desc='Stator slot height')
        self.add_input('h_r',val=0.0, units ='m', desc='Rotor slot height')
        
        self.add_input('machine_rating',val=0.0, units ='W', desc='Machine rating')
        self.add_input('n_nom',val=0.0, units ='rpm', desc='rated speed')
        
        self.add_input('B_symax',val=0.0, desc='Peak Stator Yoke flux density B_symax')
        self.add_input('I_0', val=0.0, units='A', desc='no-load excitation current')
        
        self.add_input('shaft_cm',val= np.array([0.0, 0.0, 0.0]), units='m', desc='Main Shaft CM')
        self.add_input('shaft_length',val=0.0, units='m', desc='main shaft length')
        self.add_input('Gearbox_efficiency',val=0.0, desc='Gearbox efficiency')
        
        # Material properties
        self.add_input('rho_Fe',val=0.0,units='kg*m**-3', desc='Magnetic Steel density ')
        self.add_input('rho_Copper',val=0.0,units='kg*m**-3', desc='Copper density ')
        
        # DFIG generator design output
        # Magnetic loading
        self.add_output('B_g',val=0.0, desc='Peak air gap flux density B_g')
        self.add_output('B_g1',val=0.0,  desc='air gap flux density fundamental ')
        self.add_output('B_rymax',val=0.0,  desc='maximum flux density in rotor yoke')
        self.add_output('B_tsmax',val=0.0,  desc='maximum tooth flux density in stator')
        self.add_output('B_trmax',val=0.0, desc='maximum tooth flux density in rotor')
        
        #Stator design
        self.add_output('q1',val=0.0, desc='Slots per pole per phase')
        self.add_output('N_s',val=0.0,desc='Stator turns')
        self.add_output('S',val=0.0, desc='Stator slots')
        self.add_output('h_ys',val=0.0, desc='Stator Yoke height')
        self.add_output('b_s',val=0.0,desc='stator slot width')
        self.add_output('b_t',val=0.0, desc='stator tooth width')
        self.add_output('D_ratio',val=0.0,desc='Stator diameter ratio')
        self.add_output('A_Cuscalc',val=0.0, desc='Stator Conductor cross-section mm^2')
        self.add_output('Slot_aspect_ratio1',val=0.0, desc='Stator slot apsect ratio')
        
        #Rotor design
        self.add_output('h_yr',val=0.0, desc=' rotor yoke height')
        self.add_output('tau_p',val=0.0, desc='Pole pitch')
        self.add_output('p',val=0.0, desc='No of pole pairs')
        self.add_output('Q_r',val=0.0, desc='Rotor slots')
        self.add_output('b_r',val=0.0, desc='rotor slot width')
        self.add_output('b_trmin',val=0.0,desc='minimum tooth width')
        self.add_output('b_tr',val=0.0, desc='rotor tooth width')
        self.add_output('A_bar',val=0.0, desc='Rotor Conductor cross-section mm^2')
        self.add_output('Slot_aspect_ratio2',val=0.0, desc='Rotor slot apsect ratio')
        self.add_output('r_r',val=0.0,desc='rotor radius')
        
        # Electrical performance
        self.add_output('E_p',val=0.0, desc='Stator phase voltage')
        self.add_output('f',val=0.0, desc='Generator output frequency')
        self.add_output('I_s',val=0.0, desc='Generator output phase current')
        self.add_output('A_1' ,val=0.0, desc='Specific current loading')
        self.add_output('J_s',val=0.0, desc='Stator winding Current density')
        self.add_output('J_r',val=0.0, desc='Rotor winding Current density')
        self.add_output('R_s',val=0.0, desc='Stator resistance')
        self.add_output('R_R',val=0.0, desc='Rotor resistance')
        self.add_output('L_s',val=0.0, desc='Stator synchronising inductance')
        self.add_output('L_sm',val=0.0, desc='mutual inductance')
        
        # Structural performance
        self.add_output('TC1',val=0.0, desc='Torque constraint -stator')
        self.add_output('TC2',val=0.0, desc='Torque constraint-rotor')
        
        # Mass Outputs
        self.add_output('Copper', val=0.0, units='kg', desc='Copper Mass')
        self.add_output('Iron', val=0.0, units='kg', desc='Electrical Steel Mass')
        self.add_output('Structural_mass', val=0.0, units='kg', desc='Structural Mass')
        
        # Objective functions
        self.add_output('Mass',val=0.0, desc='Actual mass')
        self.add_output('K_rad',val=0.0, desc='Stack length ratio')
        self.add_output('Losses',val=0.0, desc='Total loss')
        self.add_output('gen_eff',val=0.0, desc='Generator efficiency')
        
        # Other parameters
        self.add_output('S_N',val=0.0,desc='Slip')
        self.add_output('D_ratio_UL',val=0.0, desc='Dia ratio upper limit')
        self.add_output('D_ratio_LL',val=0.0, desc='Dia ratio Lower limit')
        self.add_output('K_rad_UL',val=0.0, desc='Aspect ratio upper limit')
        self.add_output('K_rad_LL',val=0.0, desc='Aspect ratio Lower limit')
        
        self.add_output('Overall_eff',val=0.0, desc='Overall drivetrain efficiency')
        self.add_output('I',val=np.array([0.0, 0.0, 0.0]),desc='Moments of Inertia for the component [Ixx, Iyy, Izz] around its center of mass')
        self.add_output('cm', val=np.array([0.0, 0.0, 0.0]),desc='COM [x,y,z]')
                
        #self.declare_partials('*', '*', method='fd', form='central', step=1e-6)
        
    def compute(self, inputs, outputs):
        #Create internal variables based on inputs
        r_s                  = inputs['r_s']
        l_s                  = inputs['l_s']
        h_s                  = inputs['h_s']
        h_r                  = inputs['h_r']
        machine_rating       = inputs['machine_rating']
        n_nom                = inputs['n_nom']
        Gearbox_efficiency   = inputs['Gearbox_efficiency']
        I_0                  = inputs['I_0']
        rho_Fe               = inputs['rho_Fe']
        rho_Copper           = inputs['rho_Copper']
        B_symax              = inputs['B_symax']
        shaft_cm     = inputs['shaft_cm']
        shaft_length = inputs['shaft_length']
        
        #Assign values to universal constants
        g1    = 9.81    # m/s^2 acceleration due to gravity
        sigma = 21.5e3  # shear stress 
        mu_0  = pi*4e-7 # permeability of free space
        cofi  = 0.9     # power factor
        h_w   = 0.005   # wedge height
        m     = 3       # Number of phases
        
        #Assign values to design constants
        q1        = 6                         # Number of slots per pole per phase
        b_s_tau_s = 0.45                      # Stator Slot width/Slot pitch ratio
        b_r_tau_r = 0.45                      # Rotor Slot width/Slot pitch ratio
        S_N       = -0.002                    # Slip
        y_tau_p   = 12./15                    # Coil span/pole pitch
        freq      = 60                  # frequency in Hz
        
        k_y1      = sin(pi*0.5*y_tau_p)       # winding Chording factor
        k_q1      = sin(pi/6)/q1/sin(pi/6/q1) # # zone factor
        k_wd      = k_y1*k_q1                 # Calcuating winding factor
        P_Fe0h    = 4                         #specific hysteresis losses W/kg @ 1.5 T @50 Hz
        P_Fe0e    = 1                         #specific hysteresis losses W/kg @ 1.5 T @50 Hz
        rho_Cu    = 1.8*10**(-8)*1.4          # Copper resistivity
        
        n_1       = n_nom/(1-S_N)             # actual rotor speed
        
        dia       = 2*r_s                       # air gap diameter
        p         = 3                    # number of pole pairs
        K_rad     = l_s/dia              # Aspect ratio
        K_rad_LL   = 0.5                 # lower limit on aspect ratio
        K_rad_UL   = 1.5                 # upper limit on aspect ratio
        
        # Calculating air gap length
        g          =(0.1+0.012*(machine_rating)**(1./3))*1e-3
        r_r        =r_s-g                         #rotor radius
        
        tau_p      =pi*dia/2/p              # Calculating pole pitch
        S          =2*m*p*q1                      # Calculating Stator slots
        
        tau_s      =tau_p/m/q1              # Stator slot pitch
        N_slots_pp =S/(m*p*2)              # Slots per pole per phase
        
        b_s        =b_s_tau_s*tau_s               # Calculating stator slot width
        b_so       =0.004;              #Stator slot opening wdth
        b_ro       =0.004;              #Rotor slot opening wdth
        b_t        =tau_s-b_s                     #tooth width
        
        q2         =4                         # rotor slots per pole per phase
        Q_r        =2*p*m*q2              # Calculating Rotor slots
        tau_r      =pi*(dia-2*g)/Q_r          # rotot slot pitch
        b_r        =b_r_tau_r*tau_r          # Rotor slot width
        b_tr       =tau_r-b_r              # rotor tooth width
        tau_r_min  =pi*(dia-2*(g+h_r))/Q_r
        b_trmin    =tau_r_min-b_r_tau_r*tau_r_min # minumum rotor tooth width
        
        
        # Calculating equivalent slot openings
        mu_rs       =0.005
        mu_rr       =0.005
        W_s         =(b_s/mu_rs)*1e-3   # Stator
        W_r         =(b_r/mu_rr)*1e-3   # Rotor
        
        Slot_aspect_ratio1 =h_s/b_s  # Stator slot aspect ratio
        Slot_aspect_ratio2 =h_r/b_r  # Rotor slot aspect ratio
        
        # Calculating Carter factor for stator,rotor and effective air gap length
        gamma_s = (2*W_s/g)**2/(5+2*W_s/g)
        K_Cs    =(tau_s)/(tau_s-g*gamma_s*0.5) #page 3-13 Boldea Induction machines Chapter 3
        gamma_r = (2*W_r/g)**2/(5+2*W_r/g)
        K_Cr    =(tau_r)/(tau_r-g*gamma_r*0.5) #page 3-13 Boldea Boldea Induction machines Chapter 3
        K_C     =K_Cs*K_Cr
        g_eff   =K_C*g
        
        om_m    =2*pi*n_nom/60               # mechanical frequency
        om_e    =p*om_m                   # electrical frequency
        f       =n_nom*p/60
        K_s     =0.3                     # saturation factor for Iron
        n_c     =2                       #number of conductors per coil
        a1      =2                       # number of parallel paths
        
        # Calculating stator winding turns
        N_s    = np.round(2*p*N_slots_pp*n_c/a1)
        
        # Calculating Peak flux densities
        B_g1   =mu_0*3*N_s*I_0*sqrt(2)*k_y1*k_q1/(pi*p*g_eff*(1+K_s))
        B_g    =B_g1*K_C
        B_rymax=B_symax
        
        # calculating back iron thickness
        h_ys= B_g*tau_p/(B_symax*pi)
        h_yr= h_ys
        
        d_se        =dia+2*(h_ys+h_s+h_w)  # stator outer diameter
        D_ratio=d_se/dia                        # Diameter ratio
        
        # limits for Diameter ratio depending on pole pair
        if (2*p==2):
            D_ratio_LL =1.65
            D_ratio_UL =1.69
        elif (2*p==4):
            D_ratio_LL =1.46
            D_ratio_UL =1.49
        elif (2*p==6):
            D_ratio_LL =1.37
            D_ratio_UL =1.4
        elif (2*p==8):
            D_ratio_LL =1.27
            D_ratio_UL =1.3
        else:
            D_ratio_LL =1.2
            D_ratio_UL =1.24
            
        # Stator slot fill factor
        if (2*r_s>2):
            K_fills=0.65
        else:
            K_fills=0.4
        
        # Stator winding length and cross-section
        l_fs=2*(0.015+y_tau_p*tau_p/2/cos(40*pi/180))+pi*(h_s) # end connection
        l_Cus = 2*N_s*(l_fs+l_s)/a1                               #shortpitch
        A_s = b_s*(h_s-h_w)                                                                         #Slot area
        A_scalc=b_s*1000*(h_s*1000-h_w*1000)                               #Conductor cross-section (mm^2)
        A_Cus = A_s*q1*p*K_fills/N_s                                             #Conductor cross-section (m^2)
        
        A_Cuscalc = A_scalc*q1*p*K_fills/N_s
        
        # Stator winding resistance
        R_s          =l_Cus*rho_Cu/A_Cus
        
        # Calculating no-load voltage
        om_s        =(n_nom)*2*pi/60                                                                 # rated angular frequency            
        P_e         =machine_rating/(1-S_N)                                         # Electrical power
        E_p    =om_s*N_s*k_wd*r_s*l_s*B_g1*sqrt(2)
        
        S_GN=(machine_rating-S_N*machine_rating)
        T_e         =p *(S_GN)/(2*pi*freq*(1-S_N))
        I_srated    =machine_rating/3/E_p/cofi
        
        #Rotor design
        k_fillr     = 0.7                                                                        #    Rotor slot fill factor
        diff        = h_r-h_w
        A_bar      = b_r*diff                                            # bar cross section
        Beta_skin   = sqrt(pi*mu_0*freq/2/rho_Cu)           #co-efficient for skin effect correction
        k_rm        = Beta_skin*h_r                    #co-efficient for skin effect correction
        J_b         = 6e+06                                                                    # Bar current density
        K_i         = 0.864
        I_b         = 2*m*N_s*k_wd*I_srated/Q_r    # bar current
        
        # Calculating bar resistance
        
        R_rb        =rho_Cu*k_rm*(l_s)/(A_bar)
        I_er        =I_b/(2*sin(pi*p/Q_r))                # End ring current
        J_er        = 0.8*J_b                                                                # End ring current density
        A_er        =I_er/J_er                                                            # End ring cross-section
        b           =h_r                                                                # End ring dimension
        a           =A_er/b                                                                 # End ring dimension
        D_er=(r_s*2-2*g)-0.003                                                    # End ring diameter
        l_er=pi*(D_er-b)/Q_r                                                        #  End ring segment length
        
        # Calculating end ring resistance
        R_re=rho_Cu*l_er/(2*A_er*(sin(pi*p/Q_r))**2)
        
        # Calculating equivalent rotor resistance
        R_R=(R_rb+R_re)*4*m*(k_wd*N_s)**2/Q_r
        
        # Calculating Rotor and Stator teeth flux density
        B_trmax = B_g*tau_r/b_trmin
        B_tsmax=tau_s*B_g/b_t
        
        # Calculating Equivalent core lengths
        l_r=l_s+(4)*g  # for axial cooling
        l_se =l_s+(2/3)*g
        K_fe=0.95           # Iron factor
        L_e=l_se *K_fe   # radial cooling
        
        # Calculating leakage inductance in  stator
        L_ssigmas=(2*mu_0*l_s*N_s**2/p/q1)*((h_s-h_w)/(3*b_s)+h_w/b_so)  #slot leakage inductance
        L_ssigmaew=(2*mu_0*l_s*N_s**2/p/q1)*0.34*q1*(l_fs-0.64*tau_p*y_tau_p)/l_s #end winding leakage inductance
        L_ssigmag=2*mu_0*l_s*N_s**2/p/q1*(5*(g*K_C/b_so)/(5+4*(g*K_C/b_so))) # tooth tip leakage inductance
        L_s=(L_ssigmas+L_ssigmaew+L_ssigmag)  # stator leakage inductance
        L_sm =6*mu_0*l_s*tau_p*(k_wd*N_s)**2/(pi**2*(p)*g_eff*(1+K_s))
        
        lambda_ei=2.3*D_er/(4*Q_r*l_s*(sin(pi*p/Q_r)**2))*log(4.7*dia/(a+2*b))
        lambda_b=h_r/3/b_r+h_w/b_ro
        L_i=pi*dia/Q_r
        
        # Calculating leakage inductance in  rotor
        L_rsl=(mu_0*l_s)*((h_r-h_w)/(3*b_r)+h_w/b_ro)  #slot leakage inductance
        L_rel=mu_0*(l_s*lambda_b+2*lambda_ei*L_i)                  #end winding leakage inductance
        L_rtl=(mu_0*l_s)*(0.9*tau_r*0.09/g_eff) # tooth tip leakage inductance
        L_rsigma=(L_rsl+L_rtl+L_rel)*4*m*(k_wd*N_s)**2/Q_r  # rotor leakage inductance
        
        # Calculating rotor current
        I_r=sqrt(-S_N*P_e/m/R_R)
        
        I_sm=E_p/(2*pi*freq*L_sm)
        # Calculating stator currents and specific current loading
        I_s=sqrt((I_r**2+I_sm**2))
        
        A_1=2*m*N_s*I_s/pi/(2*r_s)
        
        # Calculating masses of the electromagnetically active materials
        
        V_Cuss=m*l_Cus*A_Cus                                                                                # Volume of copper in stator
        V_Cusr=(Q_r*l_s*A_bar+pi*(D_er*A_er-A_er*b))    # Volume of copper in rotor
        V_Fest=(l_s*pi*((r_s+h_s)**2-r_s**2)-2*m*q1*p*b_s*h_s*l_s) # Volume of iron in stator teeth
        V_Fesy=l_s*pi*((r_s+h_s+h_ys)**2-(r_s+h_s)**2)                # Volume of iron in stator yoke
        r_r=r_s-g                                                                                            # rotor radius
        
        V_Fert=pi*l_s*(r_r**2-(r_r-h_r)**2)-2*m*q2*p*b_r*h_r*l_s # Volume of iron in rotor teeth
        V_Fery=l_s*pi*((r_r-h_r)**2-(r_r-h_r-h_yr)**2)                                         # Volume of iron in rotor yoke
        Copper=(V_Cuss+V_Cusr)*rho_Copper                                # Mass of Copper
        M_Fest=V_Fest*rho_Fe                                                                        # Mass of stator teeth
        M_Fesy=V_Fesy*rho_Fe                                                                        # Mass of stator yoke
        M_Fert=V_Fert*rho_Fe                                                                        # Mass of rotor tooth
        M_Fery=V_Fery*rho_Fe                                                                        # Mass of rotor yoke
        Iron=M_Fest+M_Fesy+M_Fert+M_Fery
        
        Active_mass=(Copper+Iron)
        L_tot=l_s
        Structural_mass=0.0001*Active_mass**2+0.8841*Active_mass-132.5
        Mass=Active_mass+Structural_mass
        
        # Calculating Losses and efficiency
        
        # 1. Copper losses
        
        K_R=1.2 # skin effect correction coefficient
        P_Cuss=m*I_s**2*R_s*K_R  # Copper loss-stator
        P_Cusr=m*I_r**2*R_R                     # Copper loss-rotor
        P_Cusnom=P_Cuss+P_Cusr                         # Copper loss-total
        
        # Iron Losses ( from Hysteresis and eddy currents)          
        P_Hyys=M_Fesy*(B_symax/1.5)**2*(P_Fe0h*om_e/(2*pi*60))         # Hysteresis losses in stator yoke
        P_Ftys=M_Fesy*(B_symax/1.5)**2*(P_Fe0e*(om_e/(2*pi*60))**2)    # Eddy losses in stator yoke
        P_Hyd=M_Fest*(B_tsmax/1.5)**2*(P_Fe0h*om_e/(2*pi*60))                    # Hysteresis losses in stator tooth
        P_Ftd=M_Fest*(B_tsmax/1.5)**2*(P_Fe0e*(om_e/(2*pi*60))**2)     # Eddy losses in stator tooth
        P_Hyyr=M_Fery*(B_rymax/1.5)**2*(P_Fe0h*abs(S_N)*om_e/(2*pi*60)) # Hysteresis losses in rotor yoke
        P_Ftyr=M_Fery*(B_rymax/1.5)**2*(P_Fe0e*(abs(S_N)*om_e/(2*pi*60))**2) # Eddy losses in rotor yoke
        P_Hydr=M_Fert*(B_trmax/1.5)**2*(P_Fe0h*abs(S_N)*om_e/(2*pi*60))            # Hysteresis losses in rotor tooth
        P_Ftdr=M_Fert*(B_trmax/1.5)**2*(P_Fe0e*(abs(S_N)*om_e/(2*pi*60))**2) # Eddy losses in rotor tooth
        
        # Calculating Additional losses
        P_add=0.5*machine_rating/100
        P_Fesnom=P_Hyys+P_Ftys+P_Hyd+P_Ftd+P_Hyyr+P_Ftyr+P_Hydr+P_Ftdr
        Losses=P_Cusnom+P_Fesnom+P_add
        gen_eff=(P_e-Losses)*100/P_e
        Overall_eff=gen_eff*Gearbox_efficiency
        
        # Calculating current densities in the stator and rotor
        J_s=I_s/A_Cuscalc
        J_r=I_r/(A_bar)/1e6
        
        # Calculating Tangential stress constraints
        TC1=T_e/(2*pi*sigma)
        TC2=r_s**2*l_s
        
        # Calculating mass moments of inertia and center of mass
        
        I = np.array([0.0, 0.0, 0.0])
        r_out=d_se*0.5
        I[0]   = (0.5*Mass*r_out**2)
        I[1]   = (0.25*Mass*r_out**2+(1/12)*Mass*l_s**2) 
        I[2]   = I[1]
        cm = np.array([0.0, 0.0, 0.0])
        cm[0]  = shaft_cm[0] + shaft_length/2. + l_s/2.
        cm[1]  = shaft_cm[1]
        cm[2]  = shaft_cm[2]
        
        outputs['B_tsmax']            = B_tsmax
        outputs['B_trmax']            = B_trmax
        outputs['B_rymax']            = B_rymax
        outputs['B_g']                = B_g
        outputs['B_g1']               = B_g1
        outputs['q1']                 = q1
        outputs['N_s']                = N_s
        outputs['S']                  = S
        outputs['h_ys']               = h_ys
        outputs['b_s']                = b_s
        outputs['b_t']                = b_t
        outputs['D_ratio']            = D_ratio
        outputs['D_ratio_UL']         = D_ratio_UL
        outputs['D_ratio_LL']         = D_ratio_LL
        outputs['A_Cuscalc']          = A_Cuscalc
        outputs['Slot_aspect_ratio1'] = Slot_aspect_ratio1
        outputs['h_yr']               = h_yr
        outputs['tau_p']              = tau_p
        outputs['p']                  = p
        outputs['Q_r']                = Q_r
        outputs['b_r']                = b_r
        outputs['b_trmin']            = b_trmin
        outputs['b_tr']               = b_tr
        outputs['r_r']                = r_r
        outputs['S_N']                = S_N
        outputs['A_bar']              = A_bar
        outputs['Slot_aspect_ratio2'] = Slot_aspect_ratio2
        outputs['E_p']                = E_p
        outputs['f']                  = f
        outputs['I_s']                = I_s
        outputs['A_1']                = A_1
        outputs['J_s']                = J_s
        outputs['J_r']                = J_r
        outputs['R_s']                = R_s
        outputs['R_R']                = R_R
        outputs['L_s']                = L_s
        outputs['L_sm']               = L_sm
        outputs['Mass']               = Mass
        outputs['K_rad']              = K_rad
        outputs['K_rad_UL']           = K_rad_UL
        outputs['K_rad_LL']           = K_rad_LL
        outputs['Losses']             = Losses
        outputs['gen_eff']            = gen_eff
        outputs['Copper']             = Copper
        outputs['Iron']               = Iron
        outputs['Structural_mass']    = Structural_mass
        outputs['TC1']                = TC1
        outputs['TC2']                = TC2
        outputs['Overall_eff']        = Overall_eff
        outputs['cm']                 = cm
        outputs['I']                  = I




import numpy as np
import numpy.testing as npt
import unittest
import wisdem.commonse.WindWaveDrag as wwd

npts = 100
myones = np.ones((npts,))

class TestDrag(unittest.TestCase):
    def setUp(self):
        self.params = {}
        self.unknowns = {}
        self.resid = None

        # variables
        self.params['U'] = 2.0 * myones
        self.params['A'] = 4.0 * myones
        self.params['p'] = 3.0 * myones
        self.params['cm'] = 1.0
        self.params['d'] = 10.0 * myones
        self.params['rho'] = 0.5
        self.params['mu'] = 1e-3
        self.params['z'] = -100.0 * myones
        self.params['beta'] = 0.0
        self.params['cd_usr'] = -1.0
        
        self.wave = wwd.CylinderWaveDrag(nPoints=npts)


    def testRegular(self):
        U   = 2.0
        A   = 4.0
        #cm  = 1.0
        r   = 5.0
        rho = 0.5
        #mu  = 1e-3

        #Re = rho*U*2*r/mu
        q  = 0.5*rho*U*U
        cd = 1.11
        area  = 2*r
        D = q*area*cd

        Fi = rho * A * np.pi * r*r
        Fp = Fi + D

        self.wave.compute(self.params, self.unknowns)

        npt.assert_equal(self.unknowns['waveLoads_Px'], Fp)
        npt.assert_equal(self.unknowns['waveLoads_Py'], 0.0)
        npt.assert_equal(self.unknowns['waveLoads_Pz'], 0.0)
        npt.assert_equal(self.unknowns['waveLoads_qdyn'], q)
        npt.assert_equal(self.unknowns['waveLoads_pt'], q + 3.0)
        npt.assert_equal(self.unknowns['waveLoads_z'], -100.0)
        npt.assert_equal(self.unknowns['waveLoads_beta'], 0.0)
        npt.assert_equal(self.unknowns['waveLoads_d'], 10.0)

    def testCDset(self):
        self.params['cd_usr'] = 2.0
        U   = 2.0
        A   = 4.0
        r   = 5.0

        rho = 0.5
        q  = 0.5*rho*U*U
        area  = 2*r
        D = q*area*2.0

        Fi = rho * A * np.pi * r*r
        Fp = Fi + D
        self.wave.compute(self.params, self.unknowns)
        npt.assert_equal(self.unknowns['waveLoads_Px'], Fp)
        
def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestDrag))
    return suite

if __name__ == '__main__':
    unittest.TextTestRunner().run(suite())



# -*- coding: utf-8 -*-
"""
bedplateSE.py

Created on Wed Sep  4 13:49:44 2019

@author: gscott

A generalization of the 2015 bedplate model (2 parallel I-beams, cast in front, steel in rear)
"""

import sys, os
import numpy as np
#import scipy as scp
#import scipy.optimize as opt
from math import pi, cos, sqrt, sin, exp, log10, log

#from wisdem.drivetrainse.drivese_utils import get_rotor_mass, get_distance_hub2mb, get_My, get_Mz, resize_for_bearings, mainshaftFlangeCalc 
#from wisdem.commonse.utilities import assembleI, unassembleI 

#--------------------------------------------

class BPelement(object):
    ''' BPelement class
        The BPelement class represents a section (along the x-axis) of the complete bedplate. Each BPelement has its own
        dimensions, material properties, and loads.
        
        The x-origin is located at the center of the tower top. Negative values are towards rotor; positive towards generator.
        
        Need to include moments from other elements
    '''
    
    #--------------------
    
    def __init__(self, element_name, material_density, material_E, material_name, nBeams, 
                 height=0.0, length=0.0, width=0.0, debug=False):
        self.element_name     = element_name
        self.material_density = material_density
        self.material_E       = material_E
        self.material_name    = material_name
        self.nBeams           = nBeams
        self.height           = height
        self.length           = length
        self.width            = width
        self.debug            = debug
        
        self.mass             = 0.0
        
        self.cm = np.zeros(3)
        self.I  = np.zeros(3)
        
        self.loads = []
        self.loadlocs = []
        # self.moments = []

        # initial I-beam dimensions in m
        self.tf = 0.01905        # flange thickness
        self.tw = 0.0127         # web thickness
        self.h0 = 0.6096         # overall height
        self.b0 = self.h0 / 2.0  # overall width
        
    #--------------------
    
    def characterize(self, stressTol, deflTol):
        # iterate over increasing size
        
        self.bi  = (self.b0 - self.tw) / 2.0
        self.hi  = self.h0 - 2.0 * self.tf
        self.I_b = self.b0 * self.h0**3 / 12.0 - 2 * self.bi * self.hi**3 / 12.0
        self.A   = self.b0 * self.h0 - 2.0 * self.bi * self.hi # cross-sectional area of beam
        self.w   = self.A * self.material_density # weight/len in kg/m

        # total mass
        self.mass = self.w * self.length
        
        # COM
        total_mass = self.mass + np.sum(self.loads)
        xcom = self.mass * (self.xpos - self.xpos) # modify if using origin NOT at xpos
        for i in range(len(self.loads)):
            xcom += self.loads[i] * self.loadlocs[i]
        self.cm[0] = xcom / total_mass
        
    #--------------------
    
    @staticmethod
    def dumphdr():
        return '   Name    matrl  xpos  xmin  xmax    L     W     H    Mass   COM\n'
    
    def dump(self):
        return '{:10s} {:5s} {:5.2f} {:5.2f} {:5.2f} {:5.2f} {:5.2f} {:5.2f} {:5.2f} {:6.2f}'.format(self.element_name, 
                self.material_name, self.xpos, self.xpos - 0.5*self.length, self.xpos + 0.5*self.length, 
                self.length, self.width, self.height, self.mass, self.cm[0])
        
#--------------------------------------------
    
class Bedplate(object):
    ''' Bedplate class
          The Bedplate class is used to represent the bedplate of a wind turbine drivetrain.
          It contains the general properties for a wind turbine component as well as additional design load and 
            dimensional attributes as listed below.
          It contains an update method to determine the mass, mass properties, and dimensions of the component.
          
        b0, h0: I-beam width, height
        tw, tf: I-beam web thickness, flange thickness
    '''

    def __init__(self, uptower_transformer=True, debug=False):

        super(Bedplate, self).__init__()

        self.uptower_transformer = uptower_transformer # Bool(iotype = 'in', desc = 'Boolean stating if transformer is uptower')

        self.debug = debug
        
        self.nBeams = 2    # number of parallel I-beams in each element
        self.nElements = 0 # number of distinct bedplate elements (along X axis)
        
        self.elements = [] # list of bedplate elements, ordered from upwind to downwind
    
        #Standard constants and material properties
        self.g = 9.81
        
        self.steelDensity = 7800    # kg/m^3
        self.castDensity  = 7100    # kg/m^3
        self.steelE = 210e9         # Young's modulus of steel     in N/m^2
        self.castE  = 169e9         # Young's modulus of cast iron in N/m^2
        self.steelStressMax = 620e6 # yield strength of alloy steel in MPa (1e6N/m^2)
        self.castStressMax  = 200e6 # yield strength of cast iron   in MPa (1e6N/m^2)

        # parameters for sizing elements
        
        self.defl_denom  = 1500.  # factor in deflection check
        self.stressTol   = 5e5
        self.deflTol     = 1e-4
        self.stress_mult = 8.     # modified to fit industry data

        
    # ------------------
        
    # functions used in bedplate sizing
    def midDeflection(self, totalLength, loadLength, load, E, I):
        ''' Eq. 2.154 - tip deflection for load applied at x (Eq. 2.66 in 2015 rpt) '''
        defl = load * loadLength**2.0 * \
            (3.0 * totalLength - loadLength) / (6.0 * E * I)
        return defl
    
    def distDeflection(self, totalLength, distWeight, E, I):
        ''' Eq. 2.155 - tip deflection for distributed load (Eq. 2.67 in 2015 rpt)'''
        defl = distWeight * totalLength**4 / (8.0 * E * I)
        return defl
    
    # ------------------
        
    def compute(self, \
                      tower_top_diameter):
        ''', gearbox_length, gearbox_location, gearbox_mass, hss_location, hss_mass, generator_location, generator_mass, \
                      lss_location, lss_mass, lss_length, mb1_cm, mb1_facewidth, mb1_mass, mb2_cm, mb2_mass, \
                      transformer_mass, transformer_cm, rotor_diameter, machine_rating, rotor_mass, rotor_bending_moment_y, rotor_force_z, \
                      flange_length, distance_hub2mb):
        '''
        
        '''Model bedplate as 2 parallel I-beams with a rear steel frame and a front cast frame
           Deflection constraints applied at each bedplate end
           Stress constraint checked at root of front and rear bedplate sections'''

        #if self.debug:
        #    sys.stderr.write('GBox loc {} mass {}\n'.format(gearbox_location, gearbox_mass))
        
        #variables
        #self.gearbox_length = gearbox_length #Float(iotype = 'in', units = 'm', desc = 'gearbox length')
        self.tower_top_diameter = tower_top_diameter #Float(iotype = 'in', units = 'm', desc = 'tower_top_diameter')
        
        # find sizes for each element
        
        self.bedplate_length = 0.0
        self.height = 0.0
        self.width = 0.0
        self.mass = 0.0
        self.totalCastMass = 0.0
        self.totalSteelMass = 0.0
        cmx = 0.0 # accumulate COM terms
        
        for element in self.elements:
            element.characterize(self.stressTol, self.deflTol)
            self.height = np.max([self.height, element.height])
            self.width = np.max([self.width, element.width])
            self.bedplate_length += element.length
            self.mass += element.mass
            if element.material_name == 'cast':
                self.totalCastMass += element.mass 
            if element.material_name == 'steel':
                self.totalSteelMass += element.mass 
            cmx += element.mass * element.xpos
            
        #self.bedplate_length = self.frontTotalLength + self.rearTotalLength
        #self.width = self.b0 + self.tower_top_diameter
        self.width += self.tower_top_diameter
        #self.height = np.max([self.frontHeight, self.rearHeight])
  
        # calculate mass properties
        cm = np.zeros(3)
        #cm[0] = (self.totalSteelMass * self.rearTotalLength/2 
        #       - self.totalCastMass  * self.frontTotalLength/2) / self.mass 
        cm[0] = cmx / self.mass
        cm[1] = 0.0
        cm[2] = -self.height/2.
        self.cm = cm
  
        self.depth = self.bedplate_length / 2.0
  
        I = np.zeros(3)
        I[0]  = self.mass * (self.width ** 2 + self.depth ** 2) / 8
        I[1]  = self.mass * (self.depth ** 2 + self.width ** 2 + (4/3) * self.bedplate_length ** 2) / 16
        I[2]  = I[1] # is this correct?
        self.I = I

        if self.debug:
            sys.stderr.write('Bedplate: mass {:.1f} cast {:.1f} steel {:.1f} L {:.1f} m H {:.1f} m W {:.1f} m\n'.format(self.mass, 
                             self.totalCastMass, self.totalSteelMass, self.bedplate_length, self.height, self.width))
            sys.stderr.write(self.elements[0].dumphdr()) #'   Name    matrl  xpos    L     W     H    Mass\n')
            for element in self.elements:
                sys.stderr.write('{}\n'.format(element.dump()))
                
            '''
            sys.stderr.write('Bedplate: frontLen {:.1f} m rearLen {:.1f} m nFront {} nRear {} \n'.format(self.frontTotalLength, 
                             self.rearTotalLength, frontCounter, rearCounter))
            sys.stderr.write('  LSS         {:5.2f} m  {:8.1f} kg\n'.format(self.lss_location, self.lss_mass))
            sys.stderr.write('  HSS         {:5.2f} m  {:8.1f} kg\n'.format(self.hss_location, self.hss_mass))
            sys.stderr.write('  Gearbox     {:5.2f} m  {:8.1f} kg\n'.format(gearbox_location, gearbox_mass))
            sys.stderr.write('  Generator   {:5.2f} m  {:8.1f} kg\n'.format(self.generator_location, self.generator_mass))
            sys.stderr.write('  Transformer {:5.2f} m  {:8.1f} kg\n'.format(self.transformer_location, self.transformer_mass))
            '''
            
        return (self.mass, self.cm, self.I, self.bedplate_length, self.height, self.width)

#%%-----------------------------

if __name__=='__main__':
    
    debug = True
    
    tower_top_diameter = 4.5
    
    bp = Bedplate(debug=debug) 

    # ------ Define bedplate elements ------
    #  This should be the only part that needs to be changed if the bedplate configuration changes
    
    bp.elements.append(BPelement('elemFront', bp.castDensity,  bp.castE,  'cast',  bp.nBeams, debug=debug))
    bp.elements.append(BPelement('elemRear',  bp.steelDensity, bp.steelE, 'steel', bp.nBeams, debug=debug))
    
    bp.elements[0].length = 3.0
    bp.elements[1].length = 5.0
    bp.elements[0].width =  2.0
    bp.elements[1].width =  2.0
    bp.elements[0].xpos =  -1.5
    bp.elements[1].xpos =   2.5
    
    bp.elements[1].loads    = [10000.,  5000., 2000.]
    bp.elements[1].loadlocs = [-2.,     0.,    1.8]  # relative to xpos
    
    # ------ END Define bedplate elements ------
        

    bp.compute(tower_top_diameter)  
     


"""
Jake Nunemaker
National Renewable Energy Lab
07/11/2019

This module contains default vessel process times.
"""


process_times = {
    # Export Cable Installation
    "onshore_construction_time": 48,  # hr
    "trench_dig_speed": 0.1,  # km/hr
    "pull_winch_speed": 5,  # km/hr
    "tow_plow_speed": 5,  # km/hr
    # Array Cable Installation
    # General Cable Installation
    "plgr_speed": 1,  # km/hr
    "cable_load_time": 6,  # hr
    "cable_prep_time": 1,  # hr
    "cable_lower_time": 1,  # hr
    "cable_pull_in_time": 5.5,  # hr
    "cable_termination_time": 5.5,  # hr
    "cable_lay_speed": 1,  # km/hr
    "cable_lay_bury_speed": 0.3,  # km/hr
    "cable_bury_speed": 0.5,  # km/hr
    "cable_splice_time": 48,  # hr
    "cable_raise_time": 0.5,  # hr
    # Offshore Substation
    "topside_fasten_time": 12,  # hr
    "topside_release_time": 2,  # hr
    "topside_attach_time": 6,  # hr
    # Monopiles
    "mono_embed_len": 30,  # m
    "mono_drive_rate": 20,  # m/hr
    "mono_fasten_time": 12,  # hr
    "mono_release_time": 3,  # hr
    "tp_fasten_time": 8,  # hr
    "tp_release_time": 2,  # hr
    "tp_bolt_time": 4,  # hr
    "grout_cure_time": 24,  # hr
    "grout_pump_time": 2,  # hr
    # Scour Protection
    "drop_rocks_time": 10,  # hr
    "load_rocks_time": 4,  # hr
    # Turbines
    "tower_section_fasten_time": 4,  # hr, applies to all sections
    "tower_section_release_time": 3,  # hr, applies to all sections
    "tower_section_attach_time": 6,  # hr, applies to all sections
    "nacelle_fasten_time": 4,  # hr
    "nacelle_release_time": 3,  # hr
    "nacelle_attach_time": 6,  # hr
    "blade_fasten_time": 1.5,  # hr
    "blade_release_time": 1,  # hr
    "blade_attach_time": 3.5,  # hr
    # Misc.
    "site_position_time": 2,  # hr
    "rov_survey_time": 1,  # hr
    "crane_reequip_time": 1,  # hr
}



import numpy as np
import numpy.testing as npt
import unittest
import wisdem.commonse.frustum as f
from wisdem.commonse import eps

myones = np.ones((100,))
rb = 4.0
rt = 2.0
t  = 0.1
h  = 3.0

class TestFrustum(unittest.TestCase):

    def testFrustumVol(self):
        
        V = np.pi/3*h * (rb**2 + rt**2 + rb * rt)

        # Test volume- scalar and vector inputs
        self.assertEqual(f.frustumVol(rb, rt, h, False), V)
        self.assertEqual(f.frustumVol(2*rb, 2*rt, h, True), V)
        npt.assert_equal(f.frustumVol(rb*myones, rt*myones, h*myones, False), V*myones)
        npt.assert_equal(f.frustumVol(2*rb*myones, 2*rt*myones, h*myones, True), V*myones)

    def testFrustumCG_solid(self):
        
        cg_solid = h/4 * (rb**2 + 3*rt**2 + 2*rb*rt) / (rb**2 + rt**2 + rb*rt)

        # Test cg of solid- scalar and vector inputs
        self.assertEqual(f.frustumCG(rb, rt, h, False), cg_solid)
        self.assertEqual(f.frustumCG(2*rb, 2*rt, h, True), cg_solid)
        npt.assert_equal(f.frustumCG(rb*myones, rt*myones, h*myones, False), cg_solid*myones)
        npt.assert_equal(f.frustumCG(2*rb*myones, 2*rt*myones, h*myones, True), cg_solid*myones)

    def testFrustum_shell(self):

        # In limit of thickness approaching radius, should recover regular formulas
        self.assertEqual(f.frustumShellVol(rb, rb, rb, h, False), f.frustumVol(rb, rb, h, False))
        self.assertEqual(f.frustumShellVol(2*rt, 2*rt, rt, h, True), f.frustumVol(rt, rt, h, False))

        self.assertEqual(f.frustumShellCG(rb, rb, rb, h, False), f.frustumCG(rb, rb, h, False))
        self.assertEqual(f.frustumShellCG(2*rt, 2*rt, rt, h, True), f.frustumCG(rt, rt, h, False))

        self.assertEqual(f.frustumShellIzz(rb, rb, rb, h, False), f.frustumIzz(rb, rb, h, False))
        self.assertEqual(f.frustumShellIzz(2*rt, 2*rt, rt, h, True), f.frustumIzz(rt, rt, h, False))

        self.assertAlmostEqual(f.frustumShellIxx(rb, rb, rb-eps, h, False), f.frustumIxx(rb, rb, h, False))
        self.assertAlmostEqual(f.frustumShellIxx(2*rt, 2*rt, rt-eps, h, True), f.frustumIxx(rt, rt, h, False))


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestFrustum))
    return suite

if __name__ == '__main__':
    unittest.TextTestRunner().run(suite())
